# ©2024, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

from enum import Enum
import logging

from pydantic import BaseModel

logger = logging.getLogger(__name__)


class OslNodeType(Enum):
    """Types of nodes that can be found on the osl_project_tree"""

    system = "system"
    actor = "actor"


class OslTreeItem(BaseModel):
    """Expected type of an item on the osl_project_tree"""

    is_root: bool
    kind: OslNodeType
    level: int
    name: str
    uid: str


class OslProjectState(Enum):
    """States that an osl project can be in on the osl project server"""

    finished = "FINISHED"
    stopped = "STOPPED"
    stop_requested = "STOP_REQUESTED"
    paused = "PAUSED"
    pause_requested = "PAUSE_REQUESTED"
    aborted = "ABORTED"
    abort_requested = "ABORT_REQUESTED"
    gently_stopped = "GENTLY_STOPPED"
    gentle_stop_requested = "GENTLE_STOP_REQUESTED"
    processing = "PROCESSING"
    idle = "IDLE"
    not_started = "NOT STARTED"
    unknown = "UNKNOWN"

    @classmethod
    def _missing_(cls, value):
        value_lower = value.lower()
        for member in cls:
            if member.value.lower() == value_lower:
                return member
        logger.warning(f"Could not convert value {value} into a valid OslProjectState.")
        return OslProjectState.unknown


class OslProjectInformation(BaseModel):
    """Expected minimum information required to extract the ProjectInformation from"""

    state: OslProjectState
    project_id: str
    name: str
    machine: str
    location: str
    working_dir: str
    user: str


class OslProjectInformationMinimumStructure(BaseModel):
    """Expected structure of any data requested from osl to extract the ProjectInformation from"""

    projects: list[OslProjectInformation]

    def get_project_information(self) -> OslProjectInformation:
        if not len(self.projects) == 1:
            raise Exception(
                "Structure to extract project information does not match expectations. \
                Did not find exactly one project."
            )
        return self.projects[0]
