# ©2024, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

from enum import Enum
from typing import ClassVar, Generic, Optional, TypeVar

from pydantic import BaseModel

from ansys.solutions.oscillator_calibration.datamodel.osl.models import OslProjectState
from ansys.solutions.oscillator_calibration.datamodel.shared.models import Alert, ProjectInformation


class WebsocketStreamListeners(Enum):
    """Enum for websocket stream listeners ids (used in the frontend)"""

    SERVER_HEALTH = "ws_server_healthy"
    PROJECT_STATE = "ws_osl_project_state"
    FULL_PROJECT_INFORMATION = "ws_full_project_status_information"
    ALERTS = "ws_alerts"
    PROJECT_INFORMATION = "ws_project_information"
    ACTOR_DATA = "ws_actor_data"


class WebsocketStreamNames(Enum):
    """Enum for websocket stream names (used in the backend)"""

    SERVER_HEALTH = "server-healthy-stream"
    PROJECT_STATE = "osl-project-state-stream"
    FULL_PROJECT_INFORMATION = "full-project-status-information-stream"
    ALERTS = "alerts-stream"
    PROJECT_INFORMATION = "project-information-stream"
    ACTOR_DATA = "actor-data-stream"


WEBSOCKET_STREAM_IDS = [entry.value for entry in WebsocketStreamListeners]


class NewDataEvent(BaseModel):
    """Event that just sends a timestamp to indicate that new data is available"""

    time_stamp: float


class AlertsEvent(BaseModel):
    """Event that is sent when an alert is raised"""

    alert: Alert = None


class HasOnDemandFlagEvent(BaseModel):
    """Base class for events that have an on_demand flag"""

    on_demand: Optional[bool] = False


class ActorDataEvent(HasOnDemandFlagEvent):
    """Event that is sent when actor data is updated"""

    updated_actors: list[str] = []


ContentT = TypeVar("ContentT")


class EventWithCorrespondingStepAttribute(BaseModel, Generic[ContentT]):
    """Base class for events that send data that is also stored in a corresponding step attribute"""

    corresponding_step_attribute_name: ClassVar[str] = None
    content: ContentT = None


class OslProjectStateEvent(HasOnDemandFlagEvent, EventWithCorrespondingStepAttribute[OslProjectState]):
    """Event that is sent when the OSL project state is updated"""

    corresponding_step_attribute_name = "osl_project_state"


class ServerHealthyEvent(EventWithCorrespondingStepAttribute[bool]):
    """Event that is sent when the server health is updated"""

    corresponding_step_attribute_name = "server_healthy"


class ProjectInformationEvent(EventWithCorrespondingStepAttribute[ProjectInformation]):
    """Event that is sent when the project information is updated"""

    corresponding_step_attribute_name = "project_information"
