# ©2024, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

from enum import Enum, auto
from typing import List

from pydantic import BaseModel
from typing_extensions import Self

from ansys.solutions.oscillator_calibration.datamodel.osl.models import OslNodeType, OslTreeItem
from ansys.solutions.oscillator_calibration.ui.utils.icons import Icon


class SystemFileType(Enum):
    OMDB = auto()
    CSV = auto()


class TreeviewItem(BaseModel):
    """A TreeviewItem contains the necessary information to
    be displayed in the Tree component and can get linked to its parent
    and children.

    Parameters
    ----------
    id: str
        unique identifier of the TreeviewItem
        (e.g. uid of the corresponding osl node)
    text: str
        text to be displayed in the Tree component for this item
        (e.g. name of the corresponding osl node)
    icon: Icon
        icon to be displayed in the Tree component for this item
    level: int
        level of this item in the tree datastructure
        (e.g. level of the corresponding osl node)
    """

    id: str
    text: str
    icon: Icon
    level: int
    children: List["TreeviewItem"] = []
    parent: "TreeviewItem" = None

    @classmethod
    def from_osl_tree_item(cls, osl_tree_item: OslTreeItem):
        icon = (
            Icon.ROOT_NODE
            if osl_tree_item.is_root
            else Icon.SYSTEM
            if osl_tree_item.kind == OslNodeType.system
            else Icon.NODE
        )
        return TreeviewItem(id=osl_tree_item.uid, text=osl_tree_item.name, icon=icon, level=osl_tree_item.level)

    def add_child(self, child: Self):
        if child.level != self.level + 1:
            raise ValueError(
                f"Child level {child.level} incompatible with parent level {self.level}. \
                            Cannot add child TreeviewItem."
            )

        self.children.append(child)
        child.parent = self

    def to_treeview_item_dict(self):
        return {
            "id": self.id,
            "text": self.text,
            "expanded": True,
            "prefixIcon": {"src": self.icon.path()},
            "prefixIconColor": "neutral-default",
            "children": [child.to_treeview_item_dict() for child in self.children],
            "readonly": True,
        }
