# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.
import shutil
import tempfile
import os
from pathlib import Path
from typing import Annotated, Optional

from ansys.optislang.core import Optislang, logging, utils
from fastapi import Body, Depends, FastAPI, HTTPException
from fastapi.responses import PlainTextResponse

app = FastAPI()


class OptislangInstance:
    def __init__(self):
        self._osl: Optional[Optislang] = None
        self._osl_working_directory: Optional[str] = None
        self._osl_port: Optional[int] = None
        self._osl_host: Optional[str] = None
        self._osl_log_path: Optional[Path] = None

    @property
    def port(self) -> Optional[int]:
        return self._osl_port

    @property
    def host(self) -> Optional[str]:
        return self._osl_host

    def _move_project_to_temp_working_dir(self, project_path: str) -> str:
        if not self._osl_working_directory:
            raise RuntimeError("Instance working directory not initialized.")
        project_file = Path(project_path)
        dest_project_file = Path(self._osl_working_directory, project_file.name)
        shutil.copyfile(project_file, dest_project_file)
        input_files = project_file.with_name("Input_Files")
        if input_files.is_dir():
            dest_input_files = dest_project_file.with_name("Input_Files")
            shutil.copytree(input_files, dest_input_files)
        return str(dest_project_file)

    def initialize(
        self,
        project_path: str,
        project_properties_file: str,
        osl_version: int,
        loglevel: str,
        log_file_path: Optional[str],
    ) -> None:
        if self._osl is None:
            self._osl_working_directory = tempfile.mkdtemp()

            if log_file_path:
                self._osl_log_path = Path(log_file_path)
            else:
                self._osl_log_path = Path(self._osl_working_directory) / "optiSLang.log"

            self._osl_log_path.open("a").close()  # Create empty file

            dest_project_file = self._move_project_to_temp_working_dir(project_path)
            self._osl = Optislang(
                project_path=dest_project_file,
                executable=utils.get_osl_exec(osl_version)[1],
                loglevel=loglevel,
                reset=True,
                auto_relocate=True,
                shutdown_on_finished=False,
                import_project_properties_file=project_properties_file,
                ini_timeout=int(os.getenv("OSL_TIMEOUT", 300)),
                env_vars={"PYTHONPATH": ""},
            )
            # Configure logging
            osl_logger = logging.OslLogger(
                loglevel=loglevel,
                log_to_file=True,
                logfile_name=self._osl_log_path,
                log_to_stdout=True,
            )
            self._osl.__logger = osl_logger.add_instance_logger(self._osl.name, self._osl, loglevel)
            self._osl.log.info("Start analysis")
            self._osl_port = self._osl.get_osl_server().get_port()
            self._osl_host = self._osl.get_osl_server().get_host()

    def shutdown(self):
        if self._osl:
            self._osl.shutdown()
            self._osl.dispose()
        if self._osl_working_directory:
            shutil.rmtree(self._osl_working_directory, ignore_errors=True)
        self._osl = None
        self._osl_working_directory = None
        self._osl_port = None
        self._osl_host = None


_optislang_instance: Optional[OptislangInstance] = None


def get_global_optislang_instance() -> OptislangInstance:
    global _optislang_instance
    if _optislang_instance is None:
        _optislang_instance = OptislangInstance()
    return _optislang_instance


GlobalOptislangInstanceDep = Annotated[OptislangInstance, Depends(get_global_optislang_instance)]


@app.post("/start")
async def start_instance(
    optislang_instance: GlobalOptislangInstanceDep,
    project_path: str = Body(...),
    project_properties_file: str = Body(...),
    osl_version: int = Body(...),
    loglevel: str = Body(...),
    log_file_path: str = Body(...),
):
    optislang_instance.initialize(
        project_path, project_properties_file, osl_version, loglevel, log_file_path
    )
    return {"host": optislang_instance.host, "port": optislang_instance.port}


@app.post("/shutdown")
async def shutdown_instance(optislang_instance: GlobalOptislangInstanceDep):
    optislang_instance.shutdown()


@app.get("/")
async def get_port(optislang_instance: GlobalOptislangInstanceDep):
    if optislang_instance.port is None:
        raise HTTPException(status_code=500, detail="Port is not available")
    return {"port": optislang_instance.port}


@app.get("/logs", response_class=PlainTextResponse)
async def get_logs(optislang_instance: GlobalOptislangInstanceDep):
    if optislang_instance._osl_log_path is None:
        raise HTTPException(status_code=500, detail="Logs are not available")
    if not optislang_instance._osl_log_path.exists():
        raise HTTPException(status_code=500, detail="Log file does not exist")
    with optislang_instance._osl_log_path.open("r") as file:
        logs = file.read()
        return logs


@app.get("/health")
async def health():
    return "healthy"
