# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.


import json
import logging
from typing import List, Optional

from ansys.saf.glow.solution import StepModel, StepSpec, instance, transaction

from ansys.solutions.oscillator_calibration.datamodel.shared.models import ActorAndHid
from ansys.solutions.oscillator_calibration.solution.optislang_manager import OptislangManager
from ansys.solutions.oscillator_calibration.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.oscillator_calibration.solution.retry_unreliable import attempt_unreliable_action


class CommandStep(StepModel):
    selected_command: Optional[str] = None

    # Pairs of uids and hids of the actor(s) for which a control command has been requested.
    selected_actors_and_hids: Optional[List[ActorAndHid]] = None

    # Stores subset of "startable" actor and hid pairs from the selected_actors_and_hids (output).
    startable_actors_and_hids: List[ActorAndHid] = []

    @transaction(
        self=StepSpec(
            download=[
                "selected_actors_and_hids",
                "selected_command",
            ],
        ),
        problem_setup_step=StepSpec(download=["osl_loglevel"]),
    )
    @instance("problem_setup_step.osl_manager", identifier="osl_manager")
    def run_command(self, problem_setup_step: ProblemSetupStep, osl_manager: OptislangManager) -> None:
        """Update the state of root or actor nodes based on the selected command in the UI."""

        if self.selected_command == "shutdown":
            osl_manager.shutdown()
            return

        logger = logging.getLogger(__name__)

        with osl_manager.instance.optislang_client(problem_setup_step.osl_loglevel) as osl:
            for element in self.selected_actors_and_hids:
                actor_uid = element["actor_uid"]
                hid = element["hid"]

                node = osl.project.root_system
                if actor_uid != osl.project.root_system.uid:
                    node = osl.project.root_system.find_node_by_uid(actor_uid, search_depth=-1)

                # When not wait_for_completion = False, this method returns None, so nothing to check
                attempt_unreliable_action(
                    node.control, None, logger, self.selected_command, hid=hid, wait_for_completion=False
                )

    @transaction(
        self=StepSpec(
            download=[
                "selected_actors_and_hids",
            ],
            upload=[
                "startable_actors_and_hids",
            ],
        ),
        problem_setup_step=StepSpec(download=["osl_loglevel"]),
    )
    @instance("problem_setup_step.osl_manager", identifier="osl_manager")
    def get_startable_actors_and_hids(
        self, problem_setup_step: ProblemSetupStep, osl_manager: OptislangManager
    ) -> None:
        """Queries if the selected actor/hid combinations are startable in optislang and returns the results."""
        logger = logging.getLogger(__name__)

        command = "ACTOR_SUPPORTS"
        feature = "CAN_START"

        startable = []

        with osl_manager.instance.optislang_client(problem_setup_step.osl_loglevel) as osl:
            osl_server = osl.get_osl_server()
            for actor_and_hid in self.selected_actors_and_hids:
                actor_uid = actor_and_hid["actor_uid"]
                hid = actor_and_hid["hid"]

                query = {"What": command, "args": {"feature": feature}, "hid": hid, "uid": actor_uid}
                query_string = json.dumps(query)

                _, result = attempt_unreliable_action(osl_server.send_command, None, logger, query_string)

                can_start = False
                if result is not None:
                    can_start = result.get(feature.lower(), False)

                if can_start:
                    startable.append(actor_and_hid)

            self.startable_actors_and_hids = startable


def execute_command(command_step: CommandStep, command: str, actors_and_hids: List[ActorAndHid]):
    """This command will run command with the given name for the actor and hid pairs given (in a single transaction).
    It will return once it has submitted the command to osl for all actor and hid pairs,
    but it does not check the status of the project or wait for the command to complete.
    """
    command_step.selected_actors_and_hids = actors_and_hids
    command_step.selected_command = command

    command_step.run_command()


def get_startable_actors_and_hids(
    command_step: CommandStep, actors_and_hids: List[ActorAndHid]
) -> Optional[List[ActorAndHid]]:
    """Retrieves a list of actor and hid pairs for which for which the Start command is supported.
    Will return None if is has not run the query due to a communication issue with Glow.
    Internally retrieves the results of the actor_supports: CAN_START query for the actor and
    hid pairs given (in a single transaction).
    """
    command_step.selected_actors_and_hids = actors_and_hids
    command_step.get_startable_actors_and_hids()

    return command_step.startable_actors_and_hids
