# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.
from contextlib import contextmanager
from typing import Any, Iterator, Optional

import httpx
import os
from ansys.optislang.core import Optislang
from ansys.saf.glow.core.instance.identification import AbstractInstanceIdentificationClient
from ansys.saf.glow.solution import InstanceManager, FileGroupReference, FileReference
from ansys.saf.glow.storage.base import AbstractStoragePath

PRODUCT_NAME = "legacy-osl"
SERVICE_NAME = "http"
OSL_TIMEOUT = int(os.getenv("OSL_TIMEOUT", 300))


class OptislangWrapperClient:
    def __init__(self, host: str, port: int):
        self._host = host
        self._port = port

    @contextmanager
    def optislang_client(self, loglevel: str = "INFO") -> Iterator[Optislang]:
        response = httpx.get(f"http://{self._host}:{self._port}")
        response.raise_for_status()
        optislang_port = response.json()["port"]
        with Optislang(host=self._host, port=optislang_port, loglevel=loglevel) as osl:
            yield osl

    def get_logs(self) -> Optional[str]:
        response = httpx.get(f"http://{self._host}:{self._port}/logs", timeout=OSL_TIMEOUT)
        response.raise_for_status()
        logs = response.content.decode()
        return logs

    def shutdown_optislang_instance(self) -> None:
        httpx.post(f"http://{self._host}:{self._port}/shutdown", timeout=OSL_TIMEOUT).raise_for_status()


class InternalOptislangManagerImpl(InstanceManager[OptislangWrapperClient]):

    def __init__(
        self,
        instance_identification: AbstractInstanceIdentificationClient,
        method_directory: Optional[AbstractStoragePath] = None,
        **kwargs: Any,
    ):
        super().__init__(instance_identification, method_directory, **kwargs)

    def initialize(
        self,
        project_path: FileReference,
        project_properties_file: FileReference,
        input_files: FileGroupReference,
        osl_version: int,
        loglevel: str,
    ):
        """Initialize and start the optiSLang server."""
        self.initialize_service(SERVICE_NAME)

        service = self._get_service()

        if not service:
            raise RuntimeError("Service not initialized.")

        self.upload_file(project_path)
        self.upload_file(project_properties_file)
        for input_file in input_files.list_files():
            self.upload_file(input_file)

        httpx.post(
            f"http://{service.host}:{service.port}/start",
            json={
                "project_path": self.absolute_path(project_path),
                "project_properties_file": self.absolute_path(project_properties_file),
                "osl_version": osl_version,
                "loglevel": loglevel,
                "log_file_path": "",
            },
            timeout=OSL_TIMEOUT
        ).raise_for_status()

    @classmethod
    def get_product_name_implement(cls) -> str:
        """Return the name of the product in the PIM environment."""
        return PRODUCT_NAME

    def implement_get_client_object(self, hostname: str, port: int) -> OptislangWrapperClient:
        """Return an instance of the custom product's client given the host and port of the shared product instance."""
        return OptislangWrapperClient(hostname, port)

    def save_state_implement(self, protected_state_directory_path: str) -> None:
        """Save the state of the product instance to the given directory."""
        pass

    def load_state_implement(self, protected_state_directory_path: str) -> None:
        """Loads the state of the product instance from the given directory."""
        pass

    def shutdown_implement(self, protected_state_directory_path: str) -> None:
        """Gracefully terminates the product instance process."""
        self.instance.shutdown_optislang_instance()
