# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

from __future__ import annotations

import json
import logging
import os
from pathlib import Path
import time
from typing import Any, Callable, Optional, Tuple, TypeVar, Union

from ansys.optislang.core import Optislang
from ansys.optislang.core.tcp.osl_server import TcpOslServer
from ansys.saf.glow.solution import (
    FileGroupReference,
    FileReference,
    StepModel,
    StepSpec,
    instance,
    long_running,
    transaction,
)
from pydantic import BaseModel, DirectoryPath

from ansys.solutions.oscillator_calibration.datamodel import datamodel
from ansys.solutions.oscillator_calibration.datamodel.osl.models import (
    OslProjectInformationMinimumStructure,
    OslProjectState,
)
from ansys.solutions.oscillator_calibration.datamodel.shared.models import Alert, AlertLevel, ProjectInformation
from ansys.solutions.oscillator_calibration.datamodel.shared.websocket_streams import (
    ActorDataEvent,
    AlertsEvent,
    NewDataEvent,
    OslProjectStateEvent,
    ProjectInformationEvent,
    ServerHealthyEvent,
    WebsocketStreamNames,
)
from ansys.solutions.oscillator_calibration.solution.optislang_manager import OptislangManager
from ansys.solutions.oscillator_calibration.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.oscillator_calibration.solution.retry_unreliable import attempt_unreliable_action
from ansys.solutions.oscillator_calibration.utilities.common_functions import check_optislang_server

logger = logging.getLogger(__name__)
T = TypeVar("T")


class ActorHidUpdate(BaseModel):
    hid: str
    actor_status_info: dict

    def create_actor_hid_result_folder(self, parent_actor_result_folder: Path) -> Path:
        hid_folder = parent_actor_result_folder / self.hid
        hid_folder.mkdir(exist_ok=True, parents=True)
        return hid_folder

    def write_data(self, parent: ActorUpdate) -> None:
        actor_data_folder = parent.get_actor_results_folder()
        hid_folder = self.create_actor_hid_result_folder(actor_data_folder)

        information_path = hid_folder / "information.json"

        information = datamodel.extract_actor_information_data(
            self.actor_status_info, parent.actor_data, parent.kind == "system"
        )

        _write_to_file(information, information_path)

        if parent.kind == "system":

            design_table_path = hid_folder / "design_table.json"
            omdb_files_path = hid_folder / "omdb_files.json"

            design_table = datamodel.extract_design_table_data(self.actor_status_info)
            omdb_data = datamodel.extract_omdb_file_data(self.actor_status_info)

            _write_to_file(design_table, design_table_path)
            _write_to_file(omdb_data, omdb_files_path)


class ActorUpdate(BaseModel):
    result_path: DirectoryPath
    actor_uid: str
    actor_data: dict
    hid_update: Optional[ActorHidUpdate] = None
    kind: str

    def get_actor_results_folder(self) -> Path:
        return self.result_path / "actor_data" / self.actor_uid

    def create_actor_result_folder(self) -> Path:
        actor_data_folder = self.get_actor_results_folder()
        actor_data_folder.mkdir(exist_ok=True, parents=True)
        return actor_data_folder

    def write_data(self) -> None:
        actor_data_folder = self.create_actor_result_folder()

        actor_path = actor_data_folder / "actor_data.json"
        actor_log_path = actor_data_folder / "log.json"
        actor_statistics_path = actor_data_folder / "statistics.json"

        actor_data = self.actor_data

        log_data = datamodel.extract_actor_log_data(self.actor_data)
        statistics_data = datamodel.extract_actor_statistics_data(self.actor_data)

        _write_to_file(actor_data, actor_path)
        _write_to_file(log_data, actor_log_path)
        _write_to_file(statistics_data, actor_statistics_path)

        if self.hid_update:
            self.hid_update.write_data(parent=self)


def _write_to_file(data: dict, file_ref: FileReference) -> None:
    serialized = json.dumps(data, indent=4)
    file_ref.write_text(serialized)


def _get_states_ids_from_states(actor_states: dict) -> Tuple[str]:
    """Get available actor states ids from actor states response."""
    if not actor_states.get("states", None):
        return tuple([])
    return tuple([state["hid"] for state in actor_states["states"]])


class MonitoringStep(StepModel):
    """Step model of the monitoring step."""

    server_healthy: bool = False

    osl_project_state: OslProjectState = OslProjectState.not_started

    actor_to_update: Optional[str] = None
    hid_to_update: Optional[str] = None

    actor_data: FileGroupReference = FileGroupReference("actor_data/**/*.json")

    project_information: ProjectInformation = ProjectInformation()

    full_project_status_information: FileReference = FileReference("full_project_status_information.json")

    auto_update_activated: bool = True

    # Methods ---------------------------------------------------------------------------------------------------------

    def _load_actor_data(
        self, osl: Optislang, actor_uid: str, hid_to_update: Optional[str] = None
    ) -> ActorUpdate:
        osl_server: TcpOslServer = osl.get_osl_server()

        success, actor_data = self._make_osl_server_request(osl_server.get_actor_info, actor_uid)

        if not success:
            return None

        success, actor_states = self._make_osl_server_request(osl_server.get_actor_states, actor_uid)

        if not success:
            return None

        actor_states_ids = _get_states_ids_from_states(actor_states)
        # states ids are expected as entry in actor_data later
        # -> need to add it now to ensure comparability of ActorUpdate objects before/after data write:
        actor_data["states_ids"] = actor_states_ids

        kind = actor_data["kind"]

        hid_update = None
        if hid_to_update is None and actor_states_ids:
            # Default to calculate the first hid in the list, if none given.
            hid_to_update = actor_states_ids[0]

        if hid_to_update:
            success, actor_status_info = self._make_osl_server_request(
                osl_server.get_actor_status_info, actor_uid, hid=hid_to_update
            )

            hid_update = ActorHidUpdate(hid=hid_to_update, actor_status_info=actor_status_info) if success else None

        actor_update = ActorUpdate(
            result_path=self.actor_data.project_path,
            actor_uid=actor_uid,
            actor_data=actor_data,
            hid_update=hid_update,
            kind=kind,
        )

        return actor_update

    @long_running
    @transaction(
        self=StepSpec(
            upload=[
                "full_project_status_information",
            ]
        ),
        problem_setup_step=StepSpec(download=["osl_loglevel"]),
    )
    @instance("problem_setup_step.osl_manager", identifier="osl_manager")
    def update_full_project_information(
        self, problem_setup_step: ProblemSetupStep, osl_manager: OptislangManager
    ) -> None:
        logger.debug("Updating full info")
        pid = os.getpid()
        logger.debug(f"Full info is running in pid: {pid}")

        previous_full_project_status_information = {}

        with osl_manager.instance.optislang_client(problem_setup_step.osl_loglevel) as osl:
            osl_server: TcpOslServer = osl.get_osl_server()

            while check_optislang_server(osl_server):
                start_time = time.monotonic()

                success, full_project_status_information = self._make_osl_server_request(
                    osl_server.get_full_project_status_info
                )

                if success:
                    if previous_full_project_status_information != full_project_status_information:
                        logger.debug("Updating full project info")
                        _write_to_file(full_project_status_information, self.full_project_status_information)
                        self.transaction.upload(["full_project_status_information"])

                        self.transaction.raise_event(
                            message=NewDataEvent(time_stamp=time.time()),
                            stream_name=WebsocketStreamNames.FULL_PROJECT_INFORMATION.value,
                        )

                        previous_full_project_status_information = full_project_status_information
                    else:
                        logger.debug("Full project info did not change.")

                else:
                    logger.debug("Error getting full project info")

                self._wait_until(start_time + 5)  # Wait five seconds after start time before requesting again

    @long_running
    @transaction(
        self=StepSpec(
            upload=[
                "osl_project_state",
                "server_healthy",
            ]
        ),
        problem_setup_step=StepSpec(download=["osl_loglevel"]),
    )
    @instance("problem_setup_step.osl_manager", identifier="osl_manager")
    def update_server_and_project_state(
        self, problem_setup_step: ProblemSetupStep, osl_manager: OptislangManager
    ) -> None:
        logger = logging.getLogger(__name__)

        with osl_manager.instance.optislang_client(problem_setup_step.osl_loglevel) as osl:
            osl_server = osl.get_osl_server()

            server_healthy = check_optislang_server(osl_server)

            while server_healthy:
                time.sleep(1.0)

                server_healthy = check_optislang_server(osl_server)
                if server_healthy != self.server_healthy:
                    msg = f"Server healthy" if server_healthy else "Server not healthy"
                    logger.debug(msg)
                    self.server_healthy = server_healthy
                    self.transaction.upload(["server_healthy"])

                    self.transaction.raise_event(
                        message=ServerHealthyEvent(content=server_healthy),
                        stream_name=WebsocketStreamNames.SERVER_HEALTH.value,
                    )

                if not server_healthy:
                    continue

                try:
                    osl_project_state = osl_server.get_project_status()

                    if osl_project_state != self.osl_project_state:
                        logger.debug(f"State change from {self.osl_project_state} to {osl_project_state}")
                        self.osl_project_state = osl_project_state
                        self.transaction.upload(["osl_project_state"])

                        self.transaction.raise_event(
                            message=OslProjectStateEvent(content=osl_project_state),
                            stream_name=WebsocketStreamNames.PROJECT_STATE.value,
                        )

                except Exception as e:
                    logger.warn("Error on getting project status:")
                    logger.warn(e)
                    pass

    @long_running
    @transaction(
        self=StepSpec(
            upload=[
                "project_information",
                "actor_data",
            ]
        ),
        problem_setup_step=StepSpec(download=["osl_loglevel", "osl_project_tree"]),
    )
    @instance("problem_setup_step.osl_manager", identifier="osl_manager")
    def update_data(self, problem_setup_step: ProblemSetupStep, osl_manager: OptislangManager) -> None:
        logger.debug("Updating data")
        pid = os.getpid()
        logger.debug(f"Update data is running in pid: {pid}")

        previous_actor_updates = {}

        with osl_manager.instance.optislang_client(problem_setup_step.osl_loglevel) as osl:
            osl_server = osl.get_osl_server()

            while check_optislang_server(osl_server):

                start_time = time.monotonic()
                success_full_project_tree, full_project_tree = self._make_osl_server_request(osl_server.get_full_project_tree)

                osl_project_information_data_structure = OslProjectInformationMinimumStructure(**full_project_tree)
                project_information = ProjectInformation.from_osl_project_information(
                    osl_project_information_data_structure.get_project_information()
                )

                if success_full_project_tree and self.project_information != project_information:
                    logger.debug(f"Updating project information.")
                    self.project_information = project_information
                    self.transaction.upload(["project_information"])

                    self.transaction.raise_event(
                        message=ProjectInformationEvent(content=project_information),
                        stream_name=WebsocketStreamNames.PROJECT_INFORMATION.value,
                    )

                osl_project_tree = problem_setup_step.osl_project_tree

                actor_updates = {}
                updated_actor_uids = []
                success_actor_data = True

                for actor_uid in [n["uid"] for n in osl_project_tree]:
                    actor_update = self._load_actor_data(osl, actor_uid)
                    if not actor_update:
                        logger.debug(f"Could not retrieve actor data for {actor_uid}")
                        success_actor_data = False
                        continue

                    actor_updates[actor_uid] = actor_update
                    logger.debug(f"Got actor data for {actor_uid}")

                    actor_data_updated = actor_update and actor_update != previous_actor_updates.get(actor_uid)
                    if actor_data_updated:
                        actor_update.write_data()
                        updated_actor_uids.append(actor_uid)

                if updated_actor_uids:
                    self.transaction.upload(["actor_data"])

                    self.transaction.raise_event(
                        message=ActorDataEvent(updated_actors=updated_actor_uids),
                        stream_name=WebsocketStreamNames.ACTOR_DATA.value,
                    )

                    logger.debug("Updated actor data")
                else:
                    logger.debug("Actor data did not change.")

                previous_actor_updates = actor_updates

                if success_full_project_tree and success_actor_data:
                    logger.debug("Done updating data")
                else:
                    logger.debug("Problem updating data")

                self._wait_until(start_time + 5)  # Wait five seconds after start time before requesting again

    @transaction(
        self=StepSpec(
            download=[
                "actor_to_update",
                "hid_to_update",
            ],
            upload=[
                "actor_data",
                "osl_project_state",
            ],
        ),
        problem_setup_step=StepSpec(download=["osl_loglevel"]),
    )
    @instance("problem_setup_step.osl_manager", identifier="osl_manager")
    def update_specific_actor_data(self, problem_setup_step: ProblemSetupStep, osl_manager: OptislangManager) -> None:
        logger.debug("Updating data")

        with osl_manager.instance.optislang_client(problem_setup_step.osl_loglevel) as osl:
            osl_server = osl.get_osl_server()
            success, osl_project_state = self._make_osl_server_request(osl_server.get_project_status)

            logger.debug(f"Got state {osl_project_state}")

            if success and osl_project_state != self.osl_project_state:
                logger.debug(f"State change from {self.osl_project_state} to {osl_project_state}")
                self.osl_project_state = osl_project_state
                self.transaction.upload(["osl_project_state"])

                self.transaction.raise_event(
                    message=OslProjectStateEvent(content=osl_project_state, on_demand=True),
                    stream_name=WebsocketStreamNames.PROJECT_STATE.value,
                )

            actor_update = self._load_actor_data(osl, self.actor_to_update, self.hid_to_update)
            if not actor_update:
                raise RuntimeError(f"Could not retrieve actor data for {self.actor_to_update}")

            actor_update.write_data()

            self.transaction.upload(["actor_data"])

            self.transaction.raise_event(
                message=ActorDataEvent(updated_actors=[self.actor_to_update], on_demand=True),
                stream_name=WebsocketStreamNames.ACTOR_DATA.value,
            )

            logger.debug(f"Got actor data for {self.actor_to_update}")

    def _on_request_error(self, request: Any, exception: Exception):
        message = "Issues detected communicating with OptiSLang. Data has been re-requested."
        logger.warn(str(exception))
        alert = Alert(time=time.time(), level=AlertLevel.WARNING, message=message, origin=request.__name__)

        self.transaction.raise_event(message=AlertsEvent(alert=alert), stream_name=WebsocketStreamNames.ALERTS.value)

    def _make_osl_server_request(self, request: Callable[..., T], *args: Any, **kwargs) -> Tuple[bool, Union[None, T]]:
        on_error = lambda exception: self._on_request_error(request, exception)
        success, response = attempt_unreliable_action(request, on_error, logger, *args, **kwargs)
        return success, response

    def _wait_until(self, until: float):
        while until > time.monotonic():
            time.sleep(0.1)
