# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Backend of the problem setup step."""

import json
import logging
import os
from pathlib import Path
import tempfile
import traceback
from typing import List

from ansys.saf.glow.solution import (
    AssetFileReference,
    BadRequestError,
    FileGroupReference,
    FileReference,
    StepModel,
    StepSpec,
    create_instance,
    instance,
    transaction,
)
from ansys.solutions.optislang.frontend_components.project_properties import (
    ProjectProperties,
    apply_placeholders_to_properties_file,
    write_properties_file,
)
import httpx

from ansys.solutions.oscillator_calibration.solution.optislang_manager import OptislangManager

logger = logging.getLogger(__name__)


class ProblemSetupStep(StepModel):
    """Step model of the problem setup step."""

    # Parameters ------------------------------------------------------------------------------------------------------

    # Frontend persistence
    project_initialized: bool = False
    analysis_locked: bool = False
    project_locked: bool = False

    osl_loglevel: str = "INFO"
    osl_project_tree: list = []
    osl_instance_started: bool = False
    placeholders: dict = {}
    registered_files: List = []
    ui_placeholders: dict = {}
    app_metadata: dict = {}

    # File storage ----------------------------------------------------------------------------------------------------

    # Inputs
    project_file: FileReference = FileReference("Problem_Setup/oscillator_calibration.opf")
    properties_file: AssetFileReference = AssetFileReference(
        relative_path="Problem_Setup/oscillator_calibration.json", encrypted=False
    )
    metadata_file: AssetFileReference = AssetFileReference(relative_path="Problem_Setup/metadata.json", encrypted=False)
    input_files: FileGroupReference = FileGroupReference("Problem_Setup/Input_Files/*.*")
    # If folder doesn't exist, it will be created later
    upload_directory: str = os.path.join(tempfile.gettempdir(), "GLOW")

    # Outputs
    working_properties_file: FileReference = FileReference("Problem_Setup/working_properties_file.json")
    osl_log_file: FileReference = FileReference("Problem_Setup/optiSLang.log")

    # Methods ---------------------------------------------------------------------------------------------------------

    @transaction(
        self=StepSpec(
            upload=[
                "project_file",
            ]
        )
    )
    def upload_bulk_files_to_project_directory(self) -> None:
        """Upload bulk files to project directory."""
        original_project_file = Path(__file__).absolute().parent.parent / "logic" / "assets" / "oscillator_calibration.opf"
        self.project_file.write_bytes(original_project_file.read_bytes())

    @transaction(self=StepSpec(download=["metadata_file"], upload=["app_metadata"]))
    def get_app_metadata(self) -> None:
        """Read OWA metadata file."""
        self.app_metadata = json.loads(self.metadata_file.read_text())

    @transaction(
        self=StepSpec(
            download=["properties_file"],
            upload=["placeholders", "registered_files"],
        )
    )
    def get_default_placeholder_values(self):
        """Get placeholder values and definitions using the ProjectProperties class."""
        pp = ProjectProperties()
        pp.read_file(self.properties_file.path)
        self.placeholders = pp._placeholders
        self.registered_files = pp._registered_files

    @transaction(
        self=StepSpec(
            download=["properties_file", "ui_placeholders"],
            upload=[
                "working_properties_file",
                "placeholders",
                "registered_files",
            ],
        )
    )
    def write_updated_properties_file(self) -> None:
        """Write updated optiSLang project properties file."""
        Path(self.working_properties_file.path).parent.mkdir(parents=True, exist_ok=True)
        properties = apply_placeholders_to_properties_file(self.ui_placeholders, self.properties_file.path)
        self.placeholders = properties["placeholders"]
        self.registered_files = properties["registered_files"]
        write_properties_file(properties, Path(self.working_properties_file.path))

    @transaction(
        self=StepSpec(
            download=["properties_file", "ui_placeholders"],
            upload=["placeholders", "registered_files"],
        )
    )
    def update_osl_placeholders_with_ui_values(self) -> None:
        """Update placeholders with values selected by the user in the UI."""
        properties = apply_placeholders_to_properties_file(self.ui_placeholders, self.properties_file.path)
        self.placeholders = properties["placeholders"]
        self.registered_files = properties["registered_files"]

    @transaction(
        self=StepSpec(
            download=[
                "input_files",
                "project_file",
                "working_properties_file",
                "osl_loglevel",
            ],
            upload=[
                "osl_instance_started",
                "osl_project_tree",
                "osl_log_file",
            ],
        )
    )
    @create_instance("osl_manager", OptislangManager)
    def start_osl_project(self, osl_manager: OptislangManager) -> None:
        """Start optiSLang and run the project."""

        # Start optiSLang instance using instance manager.
        try:
            osl_version = int(os.getenv("OSL_VERSION"))
            osl_manager.initialize(
                project_path=self.project_file,
                project_properties_file=self.working_properties_file,
                input_files=self.input_files,
                osl_version=osl_version,
                loglevel=self.osl_loglevel,
            )
        except TypeError:
            message = "Please contact your IT admin to specify an optiSLang version in the .env file and try again"
            raise BadRequestError(message)
        except httpx.HTTPStatusError:
            message = f"Cannot start optiSLang {osl_version}. Please contact your IT admin to make sure \
            that optiSLang is reachable and try again"
            raise BadRequestError(message)
        except Exception as ex:
            logger.debug("Error starting optiSLang instance - additional information:")
            logger.debug(str(ex))
            logger.debug(traceback.format_exc())
            message = "Cannot connect to the Product Instance Manager. Please contact your IT admin to make sure \
                that this component is reachable"
            raise BadRequestError(message)

        # Get optiSLang instance
        with osl_manager.instance.optislang_client(self.osl_loglevel) as osl:
            # Get project tree
            self.osl_project_tree = osl.project._get_project_tree()
            self.transaction.upload(["osl_project_tree"])
            # Start optiSLang project
            logger.info("Start Optislang analysis")
            osl.start(wait_for_started=False, wait_for_finished=False)
            logs = osl_manager.instance.get_logs()
            self.osl_log_file.write_text(logs)
        self.osl_instance_started = True

    @transaction(self=StepSpec(upload=["osl_log_file"]))
    @instance("osl_manager")
    def fetch_optislang_logs(self, osl_manager: OptislangManager) -> None:
        logs = osl_manager.instance.get_logs()
        self.osl_log_file.write_text(logs)
