# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Backend of the project files setup step."""

from enum import Enum
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, TypedDict

from ansys.optislang.core.io import RegisteredFile as OslRegisteredFile
from ansys.saf.glow.solution import FileGroupReference, FileReference, StepModel, StepSpec, instance, transaction
import pandas as pd

from ansys.solutions.oscillator_calibration.datamodel.shared.models import ActorAndHid
from ansys.solutions.oscillator_calibration.solution.monitoring_step import MonitoringStep
from ansys.solutions.oscillator_calibration.solution.optislang_manager import OptislangManager
from ansys.solutions.oscillator_calibration.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.oscillator_calibration.utilities.common_functions import convert_path, try_deserialize

PROJECT_INFO_TAG = "project-info-tag"


class FileState(Enum):
    PENDING = "PENDING"
    FINISHED = "FINISHED"
    MISSING = "MISSING"


class BasePathMode(Enum):
    WORKING_DIR_RELATIVE = "WORKING_DIR_RELATIVE"
    PROJECT_RELATIVE = "PROJECT_RELATIVE"
    PROJECT_WORKING_DIR_RELATIVE = "PROJECT_WORKING_DIR_RELATIVE"
    REFERENCE_FILES_DIR_RELATIVE = "REFERENCE_FILES_DIR_RELATIVE"
    ABSOLUTE_PATH = "ABSOLUTE_PATH"


RegisteredFile = TypedDict(
    "RegisteredFile",
    {
        "Id": str,
        "status": FileState,
        "Filename": str,
        "Comment": str,
        "size": int,
        "Last modified": str,
        "location": str,
        "tag": str,
    },
)


SplitPath = TypedDict(
    "SplitPath",
    {
        "head": str,
        "tail": str,
    },
)


LocalLocation = TypedDict(
    "LocalLocation",
    {
        "base_path_mode": BasePathMode,
        "split_path": SplitPath,
    },
)


def _get_file_details(file_path: Path) -> Tuple[int, str]:
    try:
        size = os.path.getsize(file_path)
        last_modified_timestamp = os.path.getmtime(file_path)
        last_modified = pd.to_datetime(last_modified_timestamp, unit="s").strftime("%Y-%b-%d %H:%M:%S.%f")[:-3]
    except FileNotFoundError:
        size, last_modified = 0, ""
    return size, last_modified


def _get_file_state(location: Path, project_state: str) -> FileState:
    if location.exists():
        file_state = FileState.FINISHED
    else:
        if project_state != "FINISHED":
            file_state = FileState.PENDING
        else:
            file_state = FileState.MISSING

    return file_state


def _convert_from_osl_registered_file(
    osl_file: OslRegisteredFile, project_state: str, working_directory: Path, local_location: LocalLocation
) -> RegisteredFile:
    working_directory = convert_path(working_directory)

    base_path_mode = local_location["base_path_mode"]
    tail = local_location["split_path"]["tail"]

    location = working_directory.parent / tail

    if base_path_mode == BasePathMode.ABSOLUTE_PATH.value:
        location = tail
    elif base_path_mode == BasePathMode.PROJECT_RELATIVE.value:
        location = working_directory.parent / tail
    elif base_path_mode == BasePathMode.PROJECT_WORKING_DIR_RELATIVE.value:
        location = working_directory / tail
    elif base_path_mode == BasePathMode.REFERENCE_FILES_DIR_RELATIVE.value:
        location = working_directory.with_suffix(".opr") / tail
    elif base_path_mode == BasePathMode.WORKING_DIR_RELATIVE.value:
        location = working_directory / tail

    file = RegisteredFile()
    file_state = _get_file_state(location, project_state)

    size, last_modified = _get_file_details(location)
    file["Id"] = osl_file.id
    file["status"] = file_state
    file["Filename"] = location.name
    file["Comment"] = osl_file.comment
    file["size"] = size
    file["Last modified"] = last_modified
    file["location"] = location
    file["basepathmode"] = base_path_mode
    file["tag"] = osl_file.tag

    return file


def _get_project_info_data(project_info_file: FileReference, project_state: str) -> RegisteredFile:
    project_data_file_path = project_info_file.path

    project_data_file_size, project_data_last_modified = _get_file_details(project_data_file_path)

    filename = Path(project_info_file.path).name

    # file status is set to FINISHED to enable button functionality for the file even as it gets updated
    project_state_data = {
        "Id": "project_state",
        "status": _get_file_state(Path(project_data_file_path), project_state),
        "Filename": filename,
        "Comment": "Full project state info file",
        "size": project_data_file_size,
        "Last modified": project_data_last_modified,
        "location": project_data_file_path,
        "basepathmode": BasePathMode.PROJECT_RELATIVE.value,
        "tag": "project_state",
    }

    return project_state_data


def _create_local_location(input_dict: dict) -> LocalLocation:
    local_location = LocalLocation()

    local_location["base_path_mode"] = input_dict["base_path_mode"]["value"]
    local_location["split_path"] = input_dict["split_path"]

    return local_location


def _filter_project_relative_files(
    local_locations: Dict[str, LocalLocation], osl_files: List[OslRegisteredFile]
) -> List[OslRegisteredFile]:
    project_files = []

    for osl_file in osl_files:
        if local_locations[osl_file.tag]["base_path_mode"] == BasePathMode.PROJECT_RELATIVE.value:
            project_files.append(osl_file)

    return project_files


class ResultFilesStep(StepModel):

    selected_actor_and_hid: Optional[ActorAndHid] = None

    registered_files: List[Any] = []

    registered_files_by_actor_uid: Dict[str, List[str]] = {}

    # Relates id of a file to tail and base path mode.
    local_locations: Dict[str, Any] = {}

    file_to_download: Optional[str] = None  # tag of the file to download

    # file ref to the folder where file content will be available
    downloads: FileGroupReference = FileGroupReference("downloads/*")

    @transaction(
        self=StepSpec(upload=["registered_files_by_actor_uid", "local_locations"]),
        problem_setup_step=StepSpec(download=["osl_loglevel", "working_properties_file", "osl_project_tree"]),
    )
    @instance("problem_setup_step.osl_manager", identifier="osl_manager")
    def populate_registered_files_by_actor(
        self, problem_setup_step: ProblemSetupStep, osl_manager: OptislangManager
    ) -> None:
        project_data = try_deserialize(problem_setup_step.working_properties_file)

        registered_files = project_data["registered_files"]

        self.local_locations = {f["tag"]: _create_local_location(f["local_location"]) for f in registered_files}

        file_tags_by_actor = {}
        with osl_manager.instance.optislang_client(problem_setup_step.osl_loglevel) as osl:

            osl_project_tree = problem_setup_step.osl_project_tree

            actor_uids = [n["uid"] for n in osl_project_tree if n["kind"] != "system"]

            for actor_uid in actor_uids:
                actor = osl.project.root_system.find_node_by_uid(actor_uid, search_depth=-1)

                files = actor.get_result_files()
                file_tags_by_actor[actor_uid] = [file.tag for file in files]

            project_files = osl.project.get_result_files()

            project_files = _filter_project_relative_files(self.local_locations, project_files)
            file_tags_by_actor[osl.project.root_system.uid] = [f.tag for f in project_files]

            file_tags_by_actor[osl.project.root_system.uid].append(PROJECT_INFO_TAG)

        self.registered_files_by_actor_uid = file_tags_by_actor

    @transaction(
        self=StepSpec(upload=["registered_files"], download=["local_locations"]),
        problem_setup_step=StepSpec(download=["osl_loglevel"]),
        monitoring_step=StepSpec(download=["osl_project_state", "full_project_status_information"]),
    )
    @instance("problem_setup_step.osl_manager", identifier="osl_manager")
    def populate_project_file_details(
        self, problem_setup_step: ProblemSetupStep, monitoring_step: MonitoringStep, osl_manager: OptislangManager
    ) -> None:

        with osl_manager.instance.optislang_client(problem_setup_step.osl_loglevel) as osl:

            osl_files = osl.project.get_result_files()

            osl_files = _filter_project_relative_files(self.local_locations, osl_files)

            working_directory = osl.get_working_dir()

            registered_files = [
                _convert_from_osl_registered_file(
                    osl_file, monitoring_step.osl_project_state, working_directory, self.local_locations[osl_file.tag]
                )
                for osl_file in osl_files
            ]

            project_info = _get_project_info_data(
                monitoring_step.full_project_status_information, monitoring_step.osl_project_state
            )

            registered_files.append(project_info)
            self.registered_files = registered_files

    @transaction(
        self=StepSpec(upload=["registered_files"], download=["selected_actor_and_hid", "local_locations"]),
        problem_setup_step=StepSpec(download=["osl_loglevel"]),
        monitoring_step=StepSpec(download=["osl_project_state"]),
    )
    @instance("problem_setup_step.osl_manager", identifier="osl_manager")
    def populate_actor_file_details(
        self, problem_setup_step: ProblemSetupStep, monitoring_step: MonitoringStep, osl_manager: OptislangManager
    ) -> None:

        actor_uid = self.selected_actor_and_hid["actor_uid"]
        hid = self.selected_actor_and_hid["hid"]
        with osl_manager.instance.optislang_client(problem_setup_step.osl_loglevel) as osl:
            node = osl.project.root_system.find_node_by_uid(actor_uid, search_depth=-1)
            osl_files = node.get_result_files()

            status_info = osl.osl_server.get_actor_status_info(actor_uid, hid)

            working_directory = Path(status_info["working dir"])
            registered_files = [
                _convert_from_osl_registered_file(
                    osl_file, monitoring_step.osl_project_state, working_directory, self.local_locations[osl_file.tag]
                )
                for osl_file in osl_files
            ]

            actor_registered_files = [
                file
                for file in registered_files
                if file["basepathmode"]
                in [BasePathMode.WORKING_DIR_RELATIVE.value, BasePathMode.PROJECT_WORKING_DIR_RELATIVE.value]
            ]

            self.registered_files = actor_registered_files

    @transaction(
        self=StepSpec(download=["registered_files", "file_to_download"], upload=["downloads"]),
    )
    def download(self) -> None:
        chosen_file: RegisteredFile = next(
            file for file in self.registered_files if file["tag"] == self.file_to_download
        )
        location = Path(chosen_file["location"])

        if location.exists():
            download_folder = self.downloads.project_path / "downloads"
            download_folder.mkdir(parents=True, exist_ok=True)
            content = location.read_bytes()
            downloadable_path = download_folder / chosen_file["Filename"]
            downloadable_path.write_bytes(content)
