# ©2024, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

from typing import ClassVar, List, Type, TypeVar, get_type_hints

from ansys.saf.glow._core.step_proxy import StepProxy
from ansys.saf.glow.solution import AssetFileGroupReference, AssetFileReference, FileGroupReference, FileReference
from pydantic import BaseModel

from ansys.solutions.oscillator_calibration.datamodel.osl.models import OslProjectState
from ansys.solutions.oscillator_calibration.solution.monitoring_step import MonitoringStep
from ansys.solutions.oscillator_calibration.solution.problem_setup_step import ProblemSetupStep

T = TypeVar("T", bound="Slice")


# File references and file group references would not work, since they need to be actually FileReference proxies
# to work from the UI. Creating File Reference proxies "on the fly" would rely on implementation details which
# are more likely to change when updating Glow, so for the moment let's disallow these types in the slices.
DISALLOWED_FIELD_TYPES = [AssetFileReference, AssetFileGroupReference, FileReference, FileGroupReference]


class Slice(BaseModel):
    """The slice is a part of a step model where only some of the step fields have been requested.
    The purpose is to improve performance by requesting multiple fields of interest in one network call.
    FileReferences, FileGroupReferences and their corresponding Asset types are not permitted as slice fields."""

    @classmethod
    def create(cls: Type[T], step: StepProxy) -> T:
        """Create an instance of the class, using the given step."""
        if not cls._validated:
            raise TypeError(f"Class {cls} not validated. Ensure to add the slice decorator on the class definition.")

        fields = [field for field in cls.model_fields]
        dictionary = step.to_dict(fields)  # Call to backend is done here
        slice = cls.model_validate(dictionary)
        return slice

    _validated: ClassVar[bool] = False
    """Whether this class has been properly validated against the step which it is a slice of.
    Unvalidated classes can not be created."""


def slice(step_type: Type[BaseModel]):
    def _is_subset(model: Type[T]):
        slice_hints = get_type_hints(model)
        step_hints = get_type_hints(step_type)

        for field_name in model.model_fields:
            if field_name not in step_type.model_fields:
                raise TypeError(f"Field {field_name} of {model} not in {step_type}")

            expected_type_hint = step_hints[field_name]
            actual_type_hint = slice_hints[field_name]

            if not expected_type_hint == actual_type_hint:
                raise TypeError(
                    f"Field '{field_name}' of {model} wrong type: expected {expected_type_hint}, got {actual_type_hint}"
                )

            if expected_type_hint in DISALLOWED_FIELD_TYPES:
                raise TypeError(f"Field '{field_name}' of {model} of disallowed type: {expected_type_hint}")

        model._validated = True
        return model

    return _is_subset


@slice(ProblemSetupStep)
class ProblemSetupSlice(Slice):
    analysis_locked: bool
    app_metadata: dict
    project_locked: bool
    ui_placeholders: dict


@slice(ProblemSetupStep)
class PlaceholderSlice(Slice):
    placeholders: dict
    project_locked: bool
    registered_files: List
    ui_placeholders: dict


@slice(MonitoringStep)
class ProjectStateAndServerHealthSlice(Slice):
    osl_project_state: OslProjectState
    server_healthy: bool
