# ©2024, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import logging
from pathlib import Path
from typing import Optional, Tuple

from ansys.optislang.core.io import File, FileOutputFormat
from ansys.saf.glow.solution import FileReference, StepModel, StepSpec, instance, transaction

from ansys.solutions.oscillator_calibration.datamodel.shared.models import ActorAndHid
from ansys.solutions.oscillator_calibration.solution.optislang_manager import OptislangManager
from ansys.solutions.oscillator_calibration.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.oscillator_calibration.solution.retry_unreliable import attempt_unreliable_action
from ansys.solutions.oscillator_calibration.utilities.common_functions import convert_path


class SystemFilesDownloadStep(StepModel):
    # Inputs:
    selected_actor_and_hid: Optional[ActorAndHid] = None
    working_dir: Optional[str] = None
    actor_name: Optional[str] = None

    # Outputs:
    file_content: FileReference = FileReference("downloaded")  # file ref where the file content will be available
    file_name: Optional[str] = None

    @transaction(
        self=StepSpec(
            download=[
                "selected_actor_and_hid",
                "working_dir",
                "actor_name",
            ],
            upload=[
                "file_content",
                "file_name",
            ],
        )
    )
    def download_omdb_file_content(self) -> None:
        """Downloads the latest *.omdb file from the given working directory, if any exist."""

        working_dir = Path(self.working_dir)
        working_dir = convert_path(working_dir)
        omdb_files = [File(path) for path in working_dir.glob(f"*{self.actor_name}*.omdb")]

        if omdb_files:
            omdb_file = sorted(omdb_files, key=lambda file: file.last_modified_seconds, reverse=True)[0]
            self.file_content.write_bytes(Path(omdb_file.path).read_bytes())
            self.file_name = omdb_file.filename
        else:
            logger = logging.getLogger(__name__)
            actor_uid = self.selected_actor_and_hid["actor_uid"]
            hid = self.selected_actor_and_hid["hid"]
            logger.debug(f"Download of OMDB file failed for actor {actor_uid} and hid {hid}. No OMDB files found.")

            self.file_content.write_bytes(b"")
            self.file_name = None

    @transaction(
        self=StepSpec(
            download=[
                "selected_actor_and_hid",
            ],
            upload=[
                "file_content",
                "file_name",
            ],
        ),
        problem_setup_step=StepSpec(download=["osl_loglevel"]),
    )
    @instance("problem_setup_step.osl_manager", identifier="osl_manager")
    def export_designs_csv_and_get_file_content(
        self, problem_setup_step: ProblemSetupStep, osl_manager: OptislangManager
    ) -> None:
        """Create and download the design point csv file for the given actor and hid."""

        with osl_manager.instance.optislang_client(problem_setup_step.osl_loglevel) as osl:
            node = osl.project.root_system
            actor_uid = self.selected_actor_and_hid["actor_uid"]
            hid = self.selected_actor_and_hid["hid"]
            if actor_uid != osl.project.root_system.uid:
                node = osl.project.root_system.find_node_by_uid(actor_uid, -1)

            # When not wait_for_completion = False, this method returns None, so nothing to check
            logger = logging.getLogger(__name__)
            actor_name = node.get_name()
            file_name_without_extension = f"{actor_name}_{hid}"

            directory = Path(self.file_content.path).parent
            success, csv_file = attempt_unreliable_action(
                node.save_designs_as,
                None,
                logger,
                hid=hid,
                file_name=file_name_without_extension,
                format=FileOutputFormat.CSV,
                dir=directory,
            )

            if success and csv_file:
                location = Path(csv_file.path)
                location = convert_path(location)
                self.file_content.write_bytes(location.read_bytes())

            else:
                logger.debug(
                    f"Design point csv file could not be generated for actor {actor_uid} and hid {hid}. "
                    + "No data available."
                )
                self.file_content.write_bytes(b"")

            self.file_name = file_name_without_extension + ".csv"


def retrieve_omdb_file(
    system_files_download_step: SystemFilesDownloadStep,
    actor_and_hid: ActorAndHid,
    working_dir: str,
    actor_name: str,
) -> Tuple[FileReference, str]:
    """This command retrieves the latest omdb file (if existing) in the given working_directory
    for the given actor and hid.
    """
    system_files_download_step.selected_actor_and_hid = actor_and_hid
    system_files_download_step.working_dir = working_dir
    system_files_download_step.actor_name = actor_name

    system_files_download_step.download_omdb_file_content()
    return (system_files_download_step.file_content, system_files_download_step.file_name)


def retrieve_csv_file(
    system_files_download_step: SystemFilesDownloadStep, actor_and_hid: ActorAndHid
) -> Tuple[FileReference, str]:
    """This command creates and downloads the design point csv file for the given actor and hid."""
    system_files_download_step.selected_actor_and_hid = actor_and_hid

    system_files_download_step.export_designs_csv_and_get_file_content()

    return (system_files_download_step.file_content, system_files_download_step.file_name)
