# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import sys
from typing import List

if sys.version_info < (3, 11):
    from typing_extensions import NotRequired, TypedDict
else:
    from typing import NotRequired

import dash_bootstrap_components as dbc
from dash_extensions.enrich import html

from ansys.solutions.oscillator_calibration.ui.utils.colors import ColorFilter
from ansys.solutions.oscillator_calibration.ui.utils.icons import Icon

ButtonId = TypedDict("ButtonId", {"type": str, "action": str})
ButtonOptions = TypedDict(
    "ButtonOptions",
    {
        "icon": Icon,
        "text": NotRequired[str],
        "tooltip": str,
        "disabled": NotRequired[bool],
        "id": ButtonId,
        "autofit": NotRequired[bool],
    },
)


DEFAULT_BUTTON_STYLE = {
    "display": "flex",
    "justifyContent": "center",
    "alignItems": "center",
    "fontSize": "150%",
    "color": "rgba(0, 0, 0, 1)",
    "backgroundColor": "rgba(255, 255, 255, 1)",
    "borderColor": "rgba(0, 0, 0, 1)",
    "height": "40px",
    "width": "70px",
}


def get_tooltip_target_id(action: str) -> ButtonId:
    return {"type": "action-button-tooltip-target", "action": action}


def get_tooltip_id(action: str) -> ButtonId:
    return {"type": "action-button-tooltip", "action": action}


class ButtonGroup:
    def __init__(self, options: List[ButtonOptions]):
        self.options = options
        self.buttons = self._create_button_group()

    def _create_button_group(self):
        btn_group = []
        for button_options in self.options:
            button = ButtonWithTooltip(button_options)
            btn_group.append(button)
        return btn_group


class ButtonWithTooltip(html.Div):
    def __init__(self, options: ButtonOptions):
        tooltip = options["tooltip"]
        id = options["id"]

        text = options.get("text", None)
        autofit = options.get("autofit", False)
        is_disabled = options.get("disabled", False)

        icon = options["icon"].to_dash_image(height="1em", color=ColorFilter.DARK_TEXT, align_center=True)
        text = html.Span(text, style={"display": "inline-block", "paddingLeft": "5px"}) if text else None

        button_style = dict(DEFAULT_BUTTON_STYLE)
        if autofit:
            button_style.pop("width", None)

        button = dbc.Button(
            id=id,
            children=html.Span(
                [
                    icon,
                    text,
                ]
            ),
            style=button_style,
            disabled=is_disabled,
        )

        # Make a container for the button, so that the tooltip is displayed also when the button is disabled.
        tooltip_target_id = get_tooltip_target_id(id["action"])
        button_with_tooltip = html.Div(id=tooltip_target_id, children=button)

        tooltip_id = get_tooltip_id(id["action"])
        tooltip = tooltip
        tooltip_element = dbc.Tooltip(id=tooltip_id, target=tooltip_target_id, placement="bottom", children=tooltip)

        super().__init__(
            [
                button_with_tooltip,
                tooltip_element,
            ]
        )


class ResultfilesButtons:
    @staticmethod
    def create_buttons():

        button_style = {"margin-right": "10px", "margin-bottom": "10px"}

        buttons_layout = html.Div(
            [
                dbc.Button(
                    "Download",
                    id="download-btn",
                    n_clicks=0,
                    outline=True,
                    color="dark",
                    style=button_style,
                    disabled=True,
                ),
                dbc.Button(
                    "Preview",
                    id="preview-btn",
                    n_clicks=0,
                    outline=True,
                    color="dark",
                    style=button_style,
                    disabled=True,
                ),
                dbc.Button(
                    "Copy Link",
                    id="copy-link-btn",
                    n_clicks=0,
                    outline=True,
                    color="dark",
                    style=button_style,
                    disabled=True,
                ),
            ]
        )
        return buttons_layout
