# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import logging
from typing import List

from ansys.saf.glow.client.dashclient import DashClient
from ansys.saf.glow.core.method_status import MethodStatus
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
from dash_extensions.enrich import ALL, Input, Output, State, callback, callback_context, html, no_update

from ansys.solutions.oscillator_calibration.datamodel.osl.models import OslProjectState
from ansys.solutions.oscillator_calibration.datamodel.shared.models import ActorAndHid
from ansys.solutions.oscillator_calibration.datamodel.shared.websocket_streams import (
    OslProjectStateEvent,
    ServerHealthyEvent,
    WebsocketStreamListeners,
)
from ansys.solutions.oscillator_calibration.solution.command_step import execute_command
from ansys.solutions.oscillator_calibration.solution.definition import Oscillator_CalibrationSolution
from ansys.solutions.oscillator_calibration.solution.step_slices import ProjectStateAndServerHealthSlice
from ansys.solutions.oscillator_calibration.ui.components.button_group import ButtonGroup, ButtonId, ButtonOptions
from ansys.solutions.oscillator_calibration.ui.utils.icons import Icon
from ansys.solutions.oscillator_calibration.ui.utils.websockets_helpers import get_value_from_websocket_event_or_step


def action_button_id(action: str) -> ButtonId:
    return {"type": "action-button", "action": action}


PROJECT_BUTTON_GROUP_OPTIONS: List[ButtonOptions] = [
    {
        "icon": Icon.PLAY,
        "tooltip": "Start project execution.",
        "id": action_button_id("start"),
    },
    {
        "icon": Icon.STOP,
        "tooltip": "Stop project execution.",
        "id": action_button_id("stop_gently"),
    },
    {
        "icon": Icon.ABORT,
        "tooltip": "Abort project execution.",
        "id": action_button_id("stop"),
    },
    {
        "icon": Icon.RESET,
        "tooltip": "Reset project.",
        "id": action_button_id("reset"),
    },
    {
        "icon": Icon.SHUTDOWN,
        "tooltip": "Shuts down the optiSLang server.",
        "id": action_button_id("shutdown"),
    },
]

ACTOR_BUTTON_GROUP_OPTIONS: List[ButtonOptions] = [
    {
        "icon": Icon.PLAY,
        "tooltip": "Start node execution.",
        "id": action_button_id("start"),
    },
    {
        "icon": Icon.STOP,
        "tooltip": "Stop node execution.",
        "id": action_button_id("stop_gently"),
    },
    {
        "icon": Icon.ABORT,
        "tooltip": "Abort node execution.",
        "id": action_button_id("stop"),
    },
    {
        "icon": Icon.RESET,
        "tooltip": "Reset node.",
        "id": action_button_id("reset"),
    },
]


def _set_button_disable(server_healthy: bool, project_state: OslProjectState, options: List[ButtonOptions]) -> None:
    for option in options:
        option["disabled"] = not server_healthy

        if option["id"]["action"] == "shutdown":
            option["disabled"] = (not server_healthy) or (project_state == OslProjectState.processing)


logger = logging.getLogger(__name__)

title = html.H5("Commands", className="card-title")
divider = html.Hr(className="my-2")

button_stack = dbc.Stack(
    id="button_stack",
    direction="horizontal",
    gap=1,
)

alert_style = {
    "width": "600px",
    "height": "40px",
    "display": "flex",
    "justify-content": "left",
    "align-items": "center",
}

alert = dbc.Alert(
    id="command_alert",
    dismissable=True,
    fade=True,
    is_open=False,
    style=alert_style,
)

cols = [
    dbc.Col(button_stack, width="auto"),
    dbc.Col(alert, width="auto"),
]

row = dbc.Row(cols)
card_body = dbc.CardBody(row, className="border-0 bg-transparent")
card = dbc.Card(card_body, className="border-0 bg-transparent")

confirm_shutdown_button = dbc.Button(
    children="OK",
    id="shutdown-confirmed",
    outline=True,
    color="dark",
)

cancel_shutdown_button = dbc.Button(
    children="Cancel",
    id="shutdown-canceled",
    outline=True,
    color="dark",
)

confirm_shutdown_text = (
    "Results files will no longer be available for download or preview "
    "after shutdown of the server. Continue to shutdown?"
)

confirm_shutdown = dbc.Modal(
    children=[
        dbc.ModalHeader(dbc.ModalTitle("Confirm shutdown")),
        dbc.ModalBody(children=confirm_shutdown_text),
        dbc.ModalFooter(children=[confirm_shutdown_button, cancel_shutdown_button]),
    ],
    id="confirm-shutdown",
    is_open=False,
    size="s",
    scrollable=False,
)

triggers = html.Div(
    [
        html.Div(id="trigger", style={"display": "none"}),
        html.Div(id="action-requested", style={"display": "none"}),
        html.Div(id="action-finished", style={"display": "none"}),
        html.Div(id="action-failed", style={"display": "none"}),
        html.Div(id="action-succeeded", style={"display": "none"}),
    ]
)

layout = html.Div(
    [
        title,
        divider,
        card,
        triggers,
        confirm_shutdown,
    ]
)


def _success_message(command_name: str):
    # Since we are not waiting for the command to finish we just say "sent" here rather than succeeded.

    user_friendly_command_name = command_name.replace("_", " ").title()
    return f"{user_friendly_command_name} command sent"


def _failure_message(command_name: str):
    user_friendly_command_name = command_name.replace("_", " ").title()
    return f"{user_friendly_command_name} command failed"


@callback(
    Output("command_alert", "children"),
    Output("command_alert", "color"),
    Output("command_alert", "is_open"),
    Input("action-failed", "children"),
)
def show_failure_message(action_failed_message: str):
    if not (callback_context.triggered and action_failed_message):
        raise PreventUpdate

    return action_failed_message, "warning", bool(action_failed_message)


@callback(
    Output("command_alert", "children"),
    Output("command_alert", "color"),
    Output("command_alert", "is_open"),
    Input("action-succeeded", "children"),
)
def show_success_message(action_success_message: str):
    if not (callback_context.triggered and action_success_message):
        raise PreventUpdate

    return action_success_message, "success", bool(action_success_message)


@callback(
    Output("action-failed", "children"),
    Output("action-succeeded", "children"),
    Input("action-finished", "children"),
    State("url", "pathname"),
)
def decide_success(action_finished: str, pathname: str):
    """Decide whether the command was successful or not."""
    if not (callback_context.triggered and action_finished):
        raise PreventUpdate

    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    command_step = project.steps.command_step

    command_execution_state = command_step.get_method_state("run_command").status

    alert_success = no_update
    alert_fail = no_update

    if command_execution_state == MethodStatus.Failed:
        alert_fail = _failure_message(action_finished)
    else:
        alert_success = _success_message(action_finished)

    return alert_fail, alert_success


@callback(
    Output("action-finished", "children"),
    Input("action-requested", "children"),
    State("url", "pathname"),
    State("navigation_tree", "treeItemClicked"),
    State("selected_state_dropdown", "value"),
)
def run_command(action_requested, pathname, tree_item, hid):
    """This performs actions such as stopping, restarting, and shutting down on an instance of optiSlang
    when action is requested."""

    if not (callback_context.triggered and action_requested):
        raise PreventUpdate

    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    command_step = project.steps.command_step

    actor_id = tree_item["id"]

    logger.debug(f'Running command "{action_requested}" for actor {actor_id}, hid {hid}')

    actor_and_hid = ActorAndHid(actor_uid=actor_id, hid=hid)

    execute_command(command_step, action_requested, [actor_and_hid])

    if action_requested != "shutdown":
        monitoring_step = project.steps.monitoring_step
        monitoring_step.actor_to_update = actor_id
        if hid:
            monitoring_step.hid_to_update = hid
        else:
            monitoring_step.hid_to_update = None
        monitoring_step.update_specific_actor_data()

    logger.debug(f'Finished command "{action_requested}" for actor {actor_id}')
    return action_requested


@callback(
    Output("button_stack", "children"),
    Input("action-succeeded", "children"),
    Input("action-failed", "children"),
    Input(WebsocketStreamListeners.SERVER_HEALTH.value, "message"),
    Input(WebsocketStreamListeners.PROJECT_STATE.value, "message"),
    State("monitoring_tabs", "active_tab"),
    State("url", "pathname"),
)
def create_or_reenable_buttons(
    action_succeeded, action_failed, server_healthy_event, project_state_event, selected_tab, pathname
):
    # Purposefully triggered on page load to create the button group
    # We use different button creation option depending on if we are at project or actor level

    # It would be nice to be able to control re-enablement of the buttons
    # with Output(action_button_id(ALL), 'disabled'), but it doesn't work,
    # since if Dash detects Output(action_button_id(ALL), 'disabled')
    # "downstream" from a callback, it will not apply the disablement when the command is triggered.

    if selected_tab == "project_summary_tab":
        logger.debug("Creating/reenabling project command buttons")

        options = PROJECT_BUTTON_GROUP_OPTIONS
    else:
        logger.debug("Creating/reenabling actor command buttons")

        options = ACTOR_BUTTON_GROUP_OPTIONS

    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    monitoring_step = project.steps.monitoring_step
    monitoring_step_slice = ProjectStateAndServerHealthSlice.create(monitoring_step)

    osl_project_state = get_value_from_websocket_event_or_step(
        project_state_event, monitoring_step_slice, OslProjectStateEvent
    )

    server_healthy = get_value_from_websocket_event_or_step(
        server_healthy_event, monitoring_step_slice, ServerHealthyEvent
    )

    _set_button_disable(server_healthy, osl_project_state, options)

    button_group = ButtonGroup(options=options)

    buttons = button_group.buttons

    return buttons


@callback(
    Output("action-requested", "children"),
    Output(action_button_id(ALL), "disabled"),
    Input(action_button_id(ALL), "n_clicks"),
)
def trigger_request(n_clicks_actions):
    """This disables all project and actor command buttons on click of a button and trigger the action
    to be requested"""

    if callback_context.triggered and not all(n_click is None for n_click in n_clicks_actions):
        action = callback_context.triggered_id["action"]
        logger.debug(f"Triggering {action}")

        if action == "shutdown":
            # Show confirm modal for shutdown
            raise PreventUpdate

        return action, [True] * len(n_clicks_actions)
    else:
        raise PreventUpdate


@callback(
    Output("confirm-shutdown", "is_open"),
    Input(action_button_id("shutdown"), "n_clicks"),
    Input("shutdown-confirmed", "n_clicks"),
    Input("shutdown-canceled", "n_clicks"),
    State("confirm-shutdown", "is_open"),
)
def trigger_shutdown_confirm_modal(n_clicks_shutdown, n_clicks_shutdown_confirmed, n_clicks_cancel, is_open):
    if not n_clicks_shutdown:
        return False

    return not is_open


@callback(
    Output("action-requested", "children"),
    Input("shutdown-confirmed", "n_clicks"),
)
def shutdown_confirmed(n_clicks_shutdown_confirmed):
    if not n_clicks_shutdown_confirmed:
        return no_update
    return "shutdown"
