# ©2024, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import logging
from typing import List

from ansys.saf.glow.client.dashclient import DashClient
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
from dash_extensions.enrich import Input, Output, State, callback, callback_context, dcc, html

from ansys.solutions.oscillator_calibration.datamodel.shared.models import ActorAndHid
from ansys.solutions.oscillator_calibration.datamodel.shared.websocket_streams import (
    ServerHealthyEvent,
    WebsocketStreamListeners,
)
from ansys.solutions.oscillator_calibration.datamodel.ui.models import SystemFileType
from ansys.solutions.oscillator_calibration.solution.definition import Oscillator_CalibrationSolution
from ansys.solutions.oscillator_calibration.solution.system_files_download_step import retrieve_csv_file, retrieve_omdb_file
from ansys.solutions.oscillator_calibration.ui.components.button_group import ButtonGroup, ButtonId, ButtonOptions
from ansys.solutions.oscillator_calibration.ui.utils.colors import ColorHex, ColorRgba
from ansys.solutions.oscillator_calibration.ui.utils.icons import Icon
from ansys.solutions.oscillator_calibration.ui.utils.websockets_helpers import (
    get_value_from_websocket_event_or_step,
    prevent_actor_data_event_update_if_actor_not_updated,
    prevent_event_triggered_update_if_auto_update_is_disabled,
)
from ansys.solutions.oscillator_calibration.utilities.common_functions import extract_dict_by_key, get_omdb_files_data


def system_file_download_button_id(action: str) -> ButtonId:
    return {"type": "system-file-download-button", "action": action}


SYSTEM_FILE_DOWNLOAD_BUTTON_GROUP_OPTIONS: List[ButtonOptions] = [
    {
        "icon": Icon.DOWNLOAD,
        "text": "Download OMDB",
        "tooltip": "Download optiSLang monitoring database file.",
        "id": system_file_download_button_id("download_omdb"),
        "autofit": True,
        "disabled": True,
    },
    {
        "icon": Icon.DOWNLOAD,
        "text": "Download CSV",
        "tooltip": "Download design table CSV file.",
        "id": system_file_download_button_id("download_csv"),
        "autofit": True,
    },
]

logger = logging.getLogger(__name__)


title = html.H5("System Files", className="card-title")
divider = html.Hr(className="my-2")

data_downloading_spinner = dcc.Loading(
    type="circle",
    fullscreen=True,
    color=ColorHex.ANSYS_ORANGE.value,
    style={
        "background-color": ColorRgba.TRANSPARENT_GREY.value,
    },
    children=[
        html.Div(id="download_done_csv", style={"display": "none"}),
        html.Div(id="download_done_omdb", style={"display": "none"}),
    ],
)

buttons = ButtonGroup(options=SYSTEM_FILE_DOWNLOAD_BUTTON_GROUP_OPTIONS).buttons
button_stack = dbc.Stack(
    buttons,
    id="button_stack_system_files_download",
    direction="horizontal",
    gap=1,
)

row = dbc.Row(
    dbc.Col(button_stack, width="auto"),
)
card_body = dbc.CardBody(row, className="border-0 bg-transparent")
card = dbc.Card(card_body, className="border-0 bg-transparent")
download_div = dcc.Download(id="download_system_files")

layout = html.Div(
    [
        title,
        divider,
        data_downloading_spinner,
        card,
        download_div,
    ]
)


@callback(
    Output("download_done_omdb", "children"),
    Output("download_system_files", "data"),
    Input(system_file_download_button_id("download_omdb"), "n_clicks"),
    State("url", "pathname"),
    State("navigation_tree", "treeItemClicked"),
    State("selected_state_dropdown", "value"),
)
def trigger_omdb_download(n_clicks, pathname, tree_item, hid):
    """Downloads the latest *.omdb file for the actor/hid."""

    if callback_context.triggered and n_clicks:
        logger.debug(f"Triggering omdb download.")

        actor_id = tree_item["id"]
        project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)

        problem_setup_step = project.steps.problem_setup_step
        actor_info = extract_dict_by_key(
            problem_setup_step.osl_project_tree, "uid", actor_id, expect_unique=True, return_index=False
        )
        actor_name = actor_info["name"]

        working_dir = get_working_dir_for_actor_and_hid(project, actor_id, hid)
        if not working_dir:
            logger.error(
                f"Download of OMDB file for actor {actor_id} and hid {hid} failed. Working directory not found."
            )
            raise PreventUpdate

        data = download_file(project, actor_id, hid, SystemFileType.OMDB, working_dir, actor_name)

        return n_clicks, data
    else:
        raise PreventUpdate


@callback(
    Output("download_done_csv", "children"),
    Output("download_system_files", "data"),
    Input(system_file_download_button_id("download_csv"), "n_clicks"),
    State("url", "pathname"),
    State("navigation_tree", "treeItemClicked"),
    State("selected_state_dropdown", "value"),
)
def trigger_csv_download(n_clicks, pathname, tree_item, hid):
    """Downloads design point data as *.csv file for the actor/hid."""

    if callback_context.triggered and n_clicks:
        logger.debug(f"Triggering csv download.")

        actor_id = tree_item["id"]
        project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)

        data = download_file(project, actor_id, hid, SystemFileType.CSV)

        return n_clicks, data
    else:
        raise PreventUpdate


@callback(
    Output(system_file_download_button_id("download_omdb"), "disabled"),
    Output(system_file_download_button_id("download_csv"), "disabled"),
    Input("selected_state_dropdown", "value"),
    Input(WebsocketStreamListeners.ACTOR_DATA.value, "message"),
    Input(WebsocketStreamListeners.SERVER_HEALTH.value, "message"),
    State("url", "pathname"),
    State("navigation_tree", "treeItemClicked"),
    State("activate_auto_update", "on"),
)
def enable_disable_buttons_based_on_file_availability_and_server_health(
    hid, actor_data_event, server_healthy_event, pathname, tree_item, auto_update
):
    """
    Disables both download buttons when the server is not healthy.
    With a healthy server:
    Enables the download_csv button always.
    Enables the download_omdb button when an *.omdb file is available for the selected actor/hid.
    Disables the download_omdb button when no *.omdb file is available for the selected actor/hid.
    """
    prevent_event_triggered_update_if_auto_update_is_disabled(callback_context, auto_update)

    actor_uid = tree_item["id"]
    prevent_actor_data_event_update_if_actor_not_updated(actor_data_event, callback_context, actor_uid)

    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    monitoring_step = project.steps.monitoring_step
    server_healthy = get_value_from_websocket_event_or_step(server_healthy_event, monitoring_step, ServerHealthyEvent)
    if server_healthy is None:
        logger.error("Could not retrieve server_healthy. Failed to enable/disable system files download buttons.")
        raise PreventUpdate

    omdb_files_data = get_omdb_files_data(monitoring_step.actor_data.list_files(), actor_uid, hid)
    omdb_files = omdb_files_data.get("Files", [])

    disable_omdb_files = not omdb_files or not server_healthy
    disable_csv_files = not server_healthy

    return disable_omdb_files, disable_csv_files


def get_working_dir_for_actor_and_hid(project, actor_id, hid):
    monitoring_step = project.steps.monitoring_step
    omdb_files_data = get_omdb_files_data(monitoring_step.actor_data.list_files(), actor_id, hid)
    working_dir = omdb_files_data.get("Working Directory", None)

    return working_dir


def download_file(
    project, actor_id, hid, requested_file_type: SystemFileType, working_dir: str = None, actor_name=None
):
    system_files_download_step = project.steps.system_files_download_step
    file_content = None
    file_name = None
    if requested_file_type == SystemFileType.OMDB:
        (file_content, file_name) = retrieve_omdb_file(
            system_files_download_step, ActorAndHid(actor_uid=actor_id, hid=hid), working_dir, actor_name
        )
    elif requested_file_type == SystemFileType.CSV:
        (file_content, file_name) = retrieve_csv_file(
            system_files_download_step, ActorAndHid(actor_uid=actor_id, hid=hid)
        )
    else:
        logger.error(f"Unknown system file type {requested_file_type} requested for download.")
        raise PreventUpdate

    if not file_content or not file_name:
        logger.error(f"Could not download {requested_file_type.name.lower()} file for actor {actor_id} and hid {hid}.")
        raise PreventUpdate

    try:
        return dcc.send_bytes(file_content.read_bytes(), file_name)
    except Exception as e:
        logger.error(f"Download of OMDB file for actor {actor_id} and hid {hid} failed. Could not read file content.")
        return None
