# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import logging
from typing import List, TypedDict

from ansys.saf.glow.client.dashclient import DashClient
import dash_bootstrap_components as dbc
from dash_extensions.enrich import Input, Output, State, callback

from ansys.solutions.oscillator_calibration.datamodel.shared.websocket_streams import (
    ServerHealthyEvent,
    WebsocketStreamListeners,
)
from ansys.solutions.oscillator_calibration.solution.definition import Oscillator_CalibrationSolution
from ansys.solutions.oscillator_calibration.ui.utils.websockets_helpers import get_value_from_websocket_event_or_step

TAB_LABEL_STYLE = {
    "color": "#000000",
    "text-color": "#000000",
}


DISABLED_LABEL_STYLE = {
    "color": "#A9A9A9",
    "text-color": "#A9A9A9",
}


TAB_ACTIVE_LABEL_STYLE = {
    "color": "#FFFFFF",
    "text-color": "#000000",
    "background-color": "#000000",
    "border-style": "solid",
    "border-color": "#000000",
}

LabeledTab = TypedDict("LabeledTab", {"label": str, "tab_id": str})

PROJECT_SUMMARY_TAB = LabeledTab(label="Project Summary", tab_id="project_summary_tab")
SCENERY_TAB = LabeledTab(label="Scenery", tab_id="scenery_tab")
DESIGN_TABLE_TAB = LabeledTab(label="Design Table", tab_id="design_table_tab")
VISUALIZATION_TAB = LabeledTab(label="Visualization", tab_id="visualization_tab")
STATUS_OVERVIEW_TAB = LabeledTab(label="Status Overview", tab_id="status_overview_tab")
PROJECT_RESULT_FILE_TAB = LabeledTab(label="Results Files", tab_id="project_result_files_tab")
ACTOR_RESULT_FILE_TAB = LabeledTab(label="Results Files", tab_id="actor_result_files_tab")
SUMMARY_TAB = LabeledTab(label="Summary", tab_id="summary_tab")

logger = logging.getLogger(__name__)


def get_root_tabs() -> List[LabeledTab]:
    return [
        PROJECT_SUMMARY_TAB,
        SCENERY_TAB,
        DESIGN_TABLE_TAB,
        VISUALIZATION_TAB,
        STATUS_OVERVIEW_TAB,
    ]


def get_actor_tabs() -> List[LabeledTab]:
    return [
        SUMMARY_TAB,
        STATUS_OVERVIEW_TAB,
        ]


def get_system_tabs() -> List[LabeledTab]:
    return [
        SUMMARY_TAB,
        DESIGN_TABLE_TAB,
        VISUALIZATION_TAB,
        STATUS_OVERVIEW_TAB,
        ]


def create_tab(labeled_tab: LabeledTab) -> dbc.Tab:
    return dbc.Tab(
        label=labeled_tab["label"],
        id = labeled_tab["tab_id"],
        tab_id=labeled_tab["tab_id"],
        label_style=TAB_LABEL_STYLE,
        active_label_style=TAB_ACTIVE_LABEL_STYLE,
        )


def update_list_of_tabs(node_info: dict, result_files_available: bool) -> List[dbc.Tab]:

    tab_infos = []

    if node_info["is_root"]:
        tab_infos = get_root_tabs()

        if result_files_available:
            tab_infos.insert(1, PROJECT_RESULT_FILE_TAB)

    elif node_info["kind"] == "system":
        tab_infos = get_system_tabs()
    elif node_info["kind"] == "actor":
        tab_infos = get_actor_tabs()

        if result_files_available:
            tab_infos.insert(1, ACTOR_RESULT_FILE_TAB)

    list_of_tabs = []
    
    if tab_infos:
        for tab_info in tab_infos:
            list_of_tabs.append(create_tab(tab_info))

    return list_of_tabs


@callback(
    Output(ACTOR_RESULT_FILE_TAB["tab_id"], "disabled"),
    Output(ACTOR_RESULT_FILE_TAB["tab_id"], "label_style"),
    Input(WebsocketStreamListeners.SERVER_HEALTH.value, "message"),
    State("url", "pathname"),
)
def enable_disable_actor_result_file_tab(server_health_event, pathname):
    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    monitoring_step = project.steps.monitoring_step

    server_healthy = get_value_from_websocket_event_or_step(server_health_event, monitoring_step, ServerHealthyEvent)

    disabled = not server_healthy

    style = DISABLED_LABEL_STYLE if disabled else TAB_LABEL_STYLE
    return disabled, style


@callback(
    Output(PROJECT_RESULT_FILE_TAB["tab_id"], "disabled"),
    Output(PROJECT_RESULT_FILE_TAB["tab_id"], "label_style"),
    Input(WebsocketStreamListeners.SERVER_HEALTH.value, "message"),
    State("url", "pathname"),
)
def enable_disable_project_result_file_tab(server_health_event, pathname):
    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    monitoring_step = project.steps.monitoring_step

    server_healthy = get_value_from_websocket_event_or_step(server_health_event, monitoring_step, ServerHealthyEvent)

    disabled = not server_healthy

    style = DISABLED_LABEL_STYLE if disabled else TAB_LABEL_STYLE
    return disabled, style