# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the monitoring step."""

from ansys.saf.glow.client.dashclient import DashClient, callback
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
import dash_daq as daq
from dash_extensions.enrich import Input, Output, State, html

from ansys.solutions.oscillator_calibration.datamodel.shared.websocket_streams import (
    WebsocketStreamListeners,
    WebsocketStreamNames,
)
from ansys.solutions.oscillator_calibration.solution.definition import Oscillator_CalibrationSolution
from ansys.solutions.oscillator_calibration.ui.components.tabs import update_list_of_tabs
from ansys.solutions.oscillator_calibration.ui.views import (
    design_table_view,
    monitoring_page_footer_view,
    project_summary_view,
    result_files_view,
    scenery_view,
    status_overview_view,
    summary_view,
    visualization_view,
)
from ansys.solutions.oscillator_calibration.utilities.common_functions import extract_dict_by_key


def layout(selected_actor: str, project: Oscillator_CalibrationSolution) -> html.Div:
    actor_info = extract_dict_by_key(
        project.steps.problem_setup_step.osl_project_tree, "uid", selected_actor, expect_unique=True, return_index=False
    )

    results_files_available = False
    registered_files_by_actor_uid = project.steps.result_files_step.registered_files_by_actor_uid
    if selected_actor in registered_files_by_actor_uid:
        if registered_files_by_actor_uid[selected_actor]:
            results_files_available = True

    list_of_tabs = update_list_of_tabs(actor_info, results_files_available)

    return html.Div(
        [
            DashClient.create_event_listener(
                project.steps.monitoring_step,
                stream_name=WebsocketStreamNames.SERVER_HEALTH.value,
                id=WebsocketStreamListeners.SERVER_HEALTH.value,
            ),
            DashClient.create_event_listener(
                project.steps.monitoring_step,
                stream_name=WebsocketStreamNames.PROJECT_STATE.value,
                id=WebsocketStreamListeners.PROJECT_STATE.value,
            ),
            DashClient.create_event_listener(
                project.steps.monitoring_step,
                stream_name=WebsocketStreamNames.FULL_PROJECT_INFORMATION.value,
                id=WebsocketStreamListeners.FULL_PROJECT_INFORMATION.value,
            ),
            DashClient.create_event_listener(
                project.steps.monitoring_step,
                stream_name=WebsocketStreamNames.PROJECT_INFORMATION.value,
                id=WebsocketStreamListeners.PROJECT_INFORMATION.value,
            ),
            DashClient.create_event_listener(
                project.steps.monitoring_step,
                stream_name=WebsocketStreamNames.ACTOR_DATA.value,
                id=WebsocketStreamListeners.ACTOR_DATA.value,
            ),
            dbc.Stack(
                [
                    daq.BooleanSwitch(
                        id="activate_auto_update",
                        on=project.steps.monitoring_step.auto_update_activated,
                        color="#FFB71B",
                        className="ms-auto mb-3",
                    ),
                    html.Div("Auto update", className="mb-3"),
                ],
                direction="horizontal",
                gap=1,
            ),
            dbc.Tabs(
                list_of_tabs,
                id="monitoring_tabs",
                active_tab=list_of_tabs[0].tab_id,
                style={
                    "color": "#000000",
                    "background-color": "#FFFFFF",
                    "text-color": "#000000",
                },
            ),
            html.Div(
                id="monitoring_page_content",
            ),
            dbc.Row(monitoring_page_footer_view.layout(selected_actor, project)),
        ]
    )


@callback(
    Output("monitoring_page_content", "children"),
    Input("monitoring_tabs", "active_tab"),
    State("url", "pathname"),
    State("navigation_tree", "treeItemClicked"),
    State("selected_state_dropdown", "value"),
)
def update_page_content(selected_tab, pathname, tree_item, actor_hid):
    """Update the page content according to the selected tab."""

    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    monitoring_step = project.steps.monitoring_step
    problem_setup_step = project.steps.problem_setup_step

    selected_actor = tree_item["id"]
    node_is_system = is_node_system(problem_setup_step, selected_actor)

    if selected_tab == "project_summary_tab":
        return project_summary_view.layout()
    elif selected_tab == "project_result_files_tab":
        return result_files_view.layout
    elif selected_tab == "actor_result_files_tab":
        return result_files_view.layout
    elif selected_tab == "summary_tab":
        return summary_view.layout(node_is_system, project, selected_actor, actor_hid)
    elif selected_tab == "scenery_tab":
        return scenery_view.layout(monitoring_step)
    elif selected_tab == "design_table_tab":
        return design_table_view.layout()
    elif selected_tab == "visualization_tab":
        return visualization_view.layout(selected_actor)
    elif selected_tab == "status_overview_tab":
        return status_overview_view.layout()


@callback(
    Output("activate_auto_update", "disabled"),
    Input("activate_auto_update", "on"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def activate_auto_update(on, pathname):
    """Enable/Disable auto update of monitoring tabs."""
    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    monitoring_step = project.steps.monitoring_step

    monitoring_step.auto_update_activated = on

    raise PreventUpdate


def is_node_system(problem_setup_step, selected_actor):
    actor_info = extract_dict_by_key(
        problem_setup_step.osl_project_tree, "uid", selected_actor, expect_unique=True, return_index=False
    )
    node_is_system = actor_info["kind"] == "system"
    return node_is_system
