# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Initialization of the frontend layout across all the steps."""


from ansys.saf.glow.client.dashclient import DashClient, callback
from ansys.saf.glow.core.method_status import MethodStatus
from ansys_web_components_dash import Tree as AwcDashTree
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
from dash_extensions.enrich import Input, Output, State, callback_context, dcc, html, no_update

from ansys.solutions.oscillator_calibration.solution.definition import Oscillator_CalibrationSolution
from ansys.solutions.oscillator_calibration.ui.pages import monitoring_page, problem_setup_page
from ansys.solutions.oscillator_calibration.ui.utils.colors import ColorRgba
from ansys.solutions.oscillator_calibration.ui.views import alerts_view, data_query_transactions_start_view, page_footer_view
from ansys.solutions.oscillator_calibration.utilities.common_functions import get_treeview_items_from_project_tree
from ansys.solutions.oscillator_calibration.utilities.deployment import is_containerized, prepend_subdomain

DEFAULT_BUTTON_STYLE = {
    "display": "flex",
    "justify-content": "center",
    "align-items": "center",
    "color": "rgba(0, 0, 0, 1)",
    "background-color": "rgba(255, 255, 255, 1)",
    "border-color": "rgba(0, 0, 0, 1)",
    "height": "40px",
}

layout = dbc.Container(
    [
        dcc.Location(id="url", refresh=False),  # represents the browser address bar and doesn't render anything
        html.Div(
            id="progress_bar",
            style={"margin-left": "35%", "margin-top": "25%"},
        ),
        dcc.Interval(id="progress_bar_update", interval=1000, n_intervals=0, disabled=True),
        dcc.Store(id="trigger_layout_display"),
        dcc.Store(id="trigger_treeview_display"),
        dcc.Store(id="trigger_body_display"),
        html.Div(id="page_layout"),
        data_query_transactions_start_view.layout(),
        alerts_view.layout,
    ],
    fluid=True,
    style={"background-color": ColorRgba.ANSYS_GREY.value},
)


@callback(
    Output("trigger_layout_display", "data"),
    Input("url", "pathname"),
)
def initialization(pathname):
    """Run methods to initialize the solution."""
    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)

    if not project.steps.problem_setup_step.project_initialized:
        project.steps.problem_setup_step.upload_bulk_files_to_project_directory()
        project.steps.problem_setup_step.get_app_metadata()
        project.steps.problem_setup_step.get_default_placeholder_values()
        project.steps.help_step.prepare_doc_file()
        project.steps.problem_setup_step.project_initialized = True

    return True


@callback(
    Output("trigger_layout_display", "data"),
    Output("progress_bar", "children"),
    Output("progress_bar", "style"),
    Output("progress_bar_update", "disabled"),
    Input("progress_bar_update", "n_intervals"),
    State("url", "pathname"),
)
def update_progress_bar(n_intervals, pathname):
    """Track status of solution initialization."""
    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step

    if not problem_setup_step.project_initialized:

        completion_rate = 0
        message = None
        methods = ["upload_bulk_files_to_project_directory", "get_app_metadata", "get_default_placeholder_values"]

        for method in methods:
            status = problem_setup_step.get_method_state(method).status
            if status == MethodStatus.Completed:
                completion_rate += 1
            elif status == MethodStatus.Running:
                message = method.replace("_", " ").capitalize()
                break

        completion_rate = round(completion_rate / len(methods) * 100)

        return (
            no_update,
            [
                dbc.Progress(
                    value=completion_rate,
                    label=f"{completion_rate} %",
                    color="rgba(255,182,35,1)",
                    style={"width": "600px", "height": "30px"},
                ),
                dbc.Label(message),
            ],
            {"margin-left": "35%", "margin-top": "25%"},
            False,
        )

    else:
        return True, [], {"display": "none"}, True


@callback(
    Output("trigger_treeview_display", "data"),
    Output("page_layout", "children"),
    Input("trigger_layout_display", "data"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def display_page_layout(trigger_layout_display, pathname):
    """Display page layout."""

    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step

    portal_ui_url = DashClient.get_portal_ui_url()
    if problem_setup_step.project_initialized:
        osl_project_tree = problem_setup_step.osl_project_tree
        treeview_items = get_treeview_items_from_project_tree(osl_project_tree)

        logo_path = "/assets/logos/ansys-solutions-horizontal-logo.png"
        if is_containerized():
            logo_path = prepend_subdomain(logo_path)

        return (
            True,
            [
                dbc.Stack(
                    [
                        html.Div(
                            [html.Img(src=logo_path, style={"width": "80%"})],
                        ),
                        html.Div(
                            [
                                dbc.Button(
                                    f"Project Name: {project.project_display_name}",
                                    id="project_name",
                                    disabled=True,
                                    style={
                                        "color": "rgba(0, 0, 0, 1)",
                                        "background-color": "rgba(255, 255, 255, 1)",
                                        "border-color": "rgba(0, 0, 0, 1)",
                                    },
                                )
                            ],
                            className="ms-auto",
                        ),
                        html.Div(
                            [
                                dbc.Button(
                                    "Help",
                                    id="open_help",
                                    style={
                                        "color": "rgba(0, 0, 0, 1)",
                                        "background-color": "rgba(242, 242, 242, 0.6)",
                                        "border-color": "rgba(0, 0, 0, 1)",
                                    },
                                ),
                            ],
                        ),
                        html.Div(
                            [
                                dbc.Button(
                                    "Back to Projects",
                                    id="return_to_portal",
                                    className="me-2",
                                    n_clicks=0,
                                    href=portal_ui_url,
                                    style={"background-color": "rgba(0, 0, 0, 1)", "border-color": "rgba(0, 0, 0, 1)"},
                                )
                                if portal_ui_url
                                else []
                            ],
                        ),
                        html.Div(
                            id="alerts_container_page",
                            style={"position": "fixed", "top": 60, "right": 10, "width": 300},
                        ),
                    ],
                    direction="horizontal",
                    gap=3,
                ),
                html.Br(),
                dbc.Row(
                    [
                        dbc.Col(
                            [
                                AwcDashTree(
                                    id="navigation_tree",
                                    multi=False,
                                    height=650,
                                    items=treeview_items,
                                    selectedItemIds=["problem_setup_step"],
                                ),
                            ],
                            width=2,
                        ),
                        dbc.Col(
                            [
                                dbc.Row(
                                    html.Div(id="body_content", style={"padding-right": "2%"}),
                                ),
                                html.Br(),
                                dbc.Row(page_footer_view.layout(portal_ui_url)),
                            ],
                            width=10,
                        ),
                    ],
                ),
                dbc.Modal(
                    [
                        dbc.ModalHeader(dbc.ModalTitle("Help")),
                        dbc.ModalBody(id="help_content"),
                        dbc.ModalFooter(dbc.Button("Close", id="close_modal")),
                    ],
                    id="help_modal",
                    is_open=False,
                    size="xl",
                    scrollable=True,
                ),
            ],
        )
    else:
        raise PreventUpdate


@callback(
    Output("body_content", "children"),
    Output("navigation_tree", "selectedItemIds"),
    Input("navigation_tree", "treeItemClicked"),
    Input("trigger_body_display", "data"),
    State("navigation_tree", "selectedItemIds"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def display_body_content(value, trigger_body_display, selectedItemIds, pathname):
    """Display body content."""

    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    if not project.steps.problem_setup_step.project_initialized:
        raise PreventUpdate

    last_selected_item = [selectedItemIds[0]]
    triggered_id = callback_context.triggered_id

    page_layout = None
    if triggered_id == "trigger_body_display" or value is None:
        page_layout = problem_setup_page.layout(project)
    elif triggered_id == "navigation_tree":
        if value["id"] is None:
            page_layout = html.H1("Welcome!")
        elif value["id"] == "problem_setup_step":
            if (
                "problem_setup_step" not in last_selected_item
            ):  # this is to minimize the number of times the page is reloaded. I noticed issues with the creation
                # of the layout when the page is reloaded too many times closely together
                page_layout = problem_setup_page.layout(project)
                last_selected_item = ["problem_setup_step"]
            else:
                raise PreventUpdate
        else:
            # Record uid of actor selected from treeview
            selected_actor = value["id"]

            # Get page layout
            page_layout = monitoring_page.layout(selected_actor, project)
            last_selected_item = ["monitoring_step"]

    return (page_layout, last_selected_item)


@callback(
    Output("navigation_tree", "items"),
    Output("trigger_body_display", "data"),
    Input("trigger_treeview_display", "data"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def display_tree_view(trigger_treeview_display, pathname):
    """Display treeview with all project nodes."""

    if not trigger_treeview_display:
        raise PreventUpdate

    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step

    if problem_setup_step.project_initialized:
        osl_project_tree = problem_setup_step.osl_project_tree
        treeview_items = get_treeview_items_from_project_tree(osl_project_tree)
        return treeview_items, True
    else:
        raise PreventUpdate


@callback(
    Output("help_modal", "is_open"),
    Output("help_content", "children"),
    Input("open_help", "n_clicks"),
    Input("close_modal", "n_clicks"),
    State("url", "pathname"),
    State("help_modal", "is_open"),
)
def open_help(n1, n2, pathname, is_open):
    if n1 or n2:
        project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
        help = project.steps.help_step.doc_file.read_bytes().decode("utf-8")

        md = dcc.Markdown(
            help,
            mathjax=True,  # Allow Latex style equations in markdown
            link_target="_blank",  # Links open in new window
        )
        return not is_open, md
    else:
        return no_update, no_update
