# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the problem setup step."""

import time

from ansys.saf.glow.client import BadRequestException
from ansys.saf.glow.client.dashclient import DashClient, callback
from ansys.saf.glow.core.method_status import MethodStatus
from ansys.solutions.optislang.frontend_components.load_sections import to_dash_sections
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
from dash_extensions.enrich import Input, Output, State, dcc, html

from ansys.solutions.oscillator_calibration.datamodel.shared.models import AlertLevel, ProjectInformation
from ansys.solutions.oscillator_calibration.solution.definition import Oscillator_CalibrationSolution
from ansys.solutions.oscillator_calibration.solution.monitoring_step import MonitoringStep
from ansys.solutions.oscillator_calibration.solution.step_slices import PlaceholderSlice, ProblemSetupSlice
from ansys.solutions.oscillator_calibration.ui.utils.colors import ColorFilter, ColorHex, ColorRgba
from ansys.solutions.oscillator_calibration.ui.utils.icons import Icon
from ansys.solutions.oscillator_calibration.ui.views import placeholder_view
from ansys.solutions.oscillator_calibration.utilities.common_functions import (
    LOG_MESSAGE_COLORS,
    get_treeview_items_from_project_tree,
)


def layout(project: Oscillator_CalibrationSolution) -> html.Div:
    """Layout of the problem setup step."""

    problem_setup_step = project.steps.problem_setup_step
    problem_setup_slice = ProblemSetupSlice.create(problem_setup_step)

    placeholder_layout = placeholder_view.layout(problem_setup_step)
    return html.Div(
        [
            # Header
            html.H1(
                problem_setup_slice.app_metadata["title"],
                style={"fontSize": "45px", "fontWeight": "normal", "lineHeight": "1.2"},
            ),
            html.P(
                problem_setup_slice.app_metadata["description"],
                style={"fontSize": "25px", "fontWeight": "normal"},
            ),
            html.Hr(className="my-2"),
            html.Br(),
            # Alerts
            dbc.Row(
                id="project-locked-alert",
                children=[
                    dbc.Alert(
                        "The selected inputs are locked. To set new inputs, create a new project.",
                        color="info",
                    )
                ],
                style={"display": "block"} if problem_setup_slice.project_locked else {"display": "none"},
            ),
            html.Div(
                id="alerts_container_problem_setup",
                style={"position": "fixed", "top": 130, "right": 10, "width": 350, "zIndex": 1000},
            ),
            html.Br(),
            placeholder_layout,
            html.Br(),
            # Start analysis
            html.P("Start Analysis", style={"font-size": "18px"}),
            html.Hr(className="my-2"),
            html.Br(),
            dbc.Row(
                [
                    html.Div(
                        [
                            dbc.Button(
                                Icon.PLAY.to_dash_image(height="1em", color=ColorFilter.WHITE),
                                id="start_analysis",
                                disabled=problem_setup_slice.analysis_locked,
                                style={
                                    "display": "flex",
                                    "justify-content": "center",
                                    "align-items": "center",
                                    "fontSize": "150%",
                                    "background-color": "rgba(0, 0, 0, 1)",
                                    "border-color": "rgba(0, 0, 0, 1)",
                                    "height": "30px",
                                },
                            ),
                            dbc.Tooltip(
                                "Start optiSLang project execution.",
                                target="start_analysis",
                            ),
                        ],
                        className="d-grid gap-2 col-5 mx-auto",
                    ),
                    dcc.Loading(
                        type="circle",
                        fullscreen=True,
                        color=ColorHex.ANSYS_ORANGE.value,
                        style={
                            "background-color": ColorRgba.TRANSPARENT_GREY.value,
                        },
                        children=html.Div(id="wait_start_analysis"),
                    ),
                    dcc.Interval(
                        id="problem_setup_alerts_update",
                        interval=1 * 3000,  # in milliseconds
                        n_intervals=0,
                    ),
                ]
            ),
        ]
    )


def _wait_for_files_ready(monitoring_step: MonitoringStep):
    default_project_information = ProjectInformation()
    while monitoring_step.project_information == default_project_information:
        time.sleep(1)


@callback(
    Output("trigger_treeview_display", "data"),
    Output("navigation_tree", "items"),
    Output("wait_start_analysis", "children"),
    Output("start_analysis", "disabled"),
    Output("osl-dash-sections", "children"),
    Output("project-locked-alert", "style"),
    Output("alerts_container_problem_setup", "children"),
    Output("problem_setup_alerts_update", "disabled"),
    Input("start_analysis", "n_clicks"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def start_analysis(n_clicks, pathname):
    """Start optiSLang project."""
    if n_clicks:
        project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
        problem_setup_step = project.steps.problem_setup_step
        monitoring_step = project.steps.monitoring_step

        ui_data = problem_setup_step.ui_placeholders
        ui_data.update({"start_analysis_requested": True})
        problem_setup_step.ui_placeholders = ui_data

        trigger_treeview_display = False
        disable_start_analysis = False

        placeholder_slice = PlaceholderSlice.create(problem_setup_step)
        osl_dash_sections = to_dash_sections(
            placeholder_slice.placeholders,
            placeholder_slice.registered_files,
            placeholder_slice.project_locked,
        )
        project_locked_alert = {"display": "none"}
        alerts_container = []
        disable_problem_setup_alerts_update = True

        # Update project properties file prior to the solve
        problem_setup_step.write_updated_properties_file()
        # Start analysis
        try:
            problem_setup_step.start_osl_project()
        except BadRequestException as e:
            alert_level = AlertLevel.ERROR
            alerts_container.append(
                dbc.Toast(
                    str(e),
                    header=alert_level.name.capitalize(),
                    is_open=True,
                    dismissable=True,
                    icon=LOG_MESSAGE_COLORS[alert_level],
                )
            )
        finally:
            method_state = problem_setup_step.get_method_state("start_osl_project")

        if method_state.status in [MethodStatus.Running, MethodStatus.Completed]:
            # Lock start analysis and ui data
            problem_setup_step.analysis_locked = True
            problem_setup_step.project_locked = True

            project.steps.result_files_step.populate_registered_files_by_actor()

            _wait_for_files_ready(monitoring_step)
            trigger_treeview_display = True
            disable_start_analysis = True

            placeholder_slice = PlaceholderSlice.create(problem_setup_step)

            osl_dash_sections = to_dash_sections(
                placeholder_slice.placeholders,
                placeholder_slice.registered_files,
                placeholder_slice.project_locked,
            )
            project_locked_alert = {"display": "none"}
            alerts_container = []
            disable_problem_setup_alerts_update = False

        osl_project_tree = problem_setup_step.osl_project_tree
        treeview_items = get_treeview_items_from_project_tree(osl_project_tree)

        return (
            trigger_treeview_display,
            treeview_items,
            True,
            disable_start_analysis,
            osl_dash_sections,
            project_locked_alert,
            alerts_container,
            disable_problem_setup_alerts_update,
        )
    else:
        raise PreventUpdate


@callback(
    Output("alerts_container_problem_setup", "children"),
    Input("problem_setup_alerts_update", "n_intervals"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def display_start_osl_project_alerts(n_intervals, pathname):
    # Get project
    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    # Get problem setup step
    problem_setup_step = project.steps.problem_setup_step

    method_state = problem_setup_step.get_method_state("start_osl_project")
    if method_state.status == MethodStatus.Failed:
        alert_level = AlertLevel.ERROR
        toast = dbc.Toast(
            method_state.exception_message,
            header=alert_level.name.capitalize(),
            is_open=True,
            dismissable=True,
            icon=LOG_MESSAGE_COLORS[alert_level],
        )

        return toast

    raise PreventUpdate
