# ©2024, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Common functions."""

import logging
from typing import Type, TypeVar

from ansys.saf.glow.solution import StepModel
from dash._callback_context import CallbackContext
from dash.exceptions import PreventUpdate

from ansys.solutions.oscillator_calibration.datamodel.shared.models import Alert
from ansys.solutions.oscillator_calibration.datamodel.shared.websocket_streams import (
    WEBSOCKET_STREAM_IDS,
    ActorDataEvent,
    AlertsEvent,
    EventWithCorrespondingStepAttribute,
    HasOnDemandFlagEvent,
    WebsocketStreamListeners,
)
from ansys.solutions.oscillator_calibration.solution.step_slices import Slice

logger = logging.getLogger(__name__)

ContentT = TypeVar("ContentT")


def get_value_from_websocket_event_or_step(
    event: dict, step: StepModel | Slice, event_type: Type[EventWithCorrespondingStepAttribute[ContentT]]
) -> ContentT:

    if not issubclass(event_type, EventWithCorrespondingStepAttribute):
        raise TypeError(
            f"Expected type {type(EventWithCorrespondingStepAttribute)}, but found type {type(event_type)}."
        )

    value = _try_to_extract_from_websocket_event_with_corresponding_step_attribute(event, event_type)

    if value is None:
        key_to_extract = event_type.corresponding_step_attribute_name
        value = getattr(step, key_to_extract, None)

        logger.debug(f"Extracted value {value} for key {key_to_extract} from step.")
    else:
        logger.debug(f"Extracted value {value} from websocket stream.")

    return value


def prevent_event_triggered_update_if_auto_update_is_disabled(
    callback_context: CallbackContext, auto_update_enabled: bool, event: dict = None, exclude_server_health: bool = True
):
    triggered_id = callback_context.triggered_id

    if exclude_server_health and triggered_id == WebsocketStreamListeners.SERVER_HEALTH.value:
        return

    if triggered_id in WEBSOCKET_STREAM_IDS:
        on_demand = _get_event_is_on_demand(event)

        if not on_demand and not auto_update_enabled:
            raise PreventUpdate


def prevent_actor_data_event_update_if_actor_not_updated(
    actor_data_event: dict, callback_context: CallbackContext, actor_id: str
):
    if callback_context.triggered_id == WebsocketStreamListeners.ACTOR_DATA.value and actor_data_event:

        message = _extract_message_from_event(actor_data_event)
        if message is None:
            raise PreventUpdate

        try:
            actor_data_event_model = ActorDataEvent.model_validate_json(message)
        except Exception as e:
            logger.debug(f"Failed to deserialize event message to expected type {type(ActorDataEvent)}: {e}")
            raise PreventUpdate

        updated_actors = actor_data_event_model.updated_actors
        if not actor_id in updated_actors:
            raise PreventUpdate


def extract_alert_from_alerts_event(alerts_event: dict) -> Alert | None:
    if alerts_event is None:
        return None

    message = _extract_message_from_event(alerts_event)
    if message is None:
        return None

    try:
        alerts_event_model = AlertsEvent.model_validate_json(message)
    except Exception as e:
        logger.debug(f"Failed to deserialize event message to expected type {type(AlertsEvent)}: {e}")
        return None

    return alerts_event_model.alert


def _try_to_extract_from_websocket_event_with_corresponding_step_attribute(
    event: dict, event_type: Type[EventWithCorrespondingStepAttribute[ContentT]]
) -> ContentT:

    if event is None:
        return None

    message = _extract_message_from_event(event)
    if message is None:
        return None

    try:
        event_model = event_type.model_validate_json(message)
    except Exception as e:
        raise ValueError(f"Failed to deserialize message to expected type {type(event_type)}: {e}")

    value = event_model.content

    return value


def _extract_message_from_event(event: dict) -> dict:
    message = event.get("data", None)
    return message


def _get_event_is_on_demand(event: dict) -> bool:

    if event is None:
        return False

    message = _extract_message_from_event(event)
    if message is None:
        return False

    try:
        event_model = HasOnDemandFlagEvent.model_validate_json(message)
    except Exception as e:
        logger.debug(f"Failed to deserialize event message to expected type {type(HasOnDemandFlagEvent)}: {e}")
        return False

    return event_model.on_demand
