# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import datetime
import time
import dash_bootstrap_components as dbc
from dash_extensions.enrich import html, Input, no_update, Output, State

from ansys.saf.glow.client.dashclient import DashClient, callback

from ansys.solutions.oscillator_calibration.datamodel.shared.websocket_streams import (
    WebsocketStreamListeners,
    WebsocketStreamNames,
)
from ansys.solutions.oscillator_calibration.solution.definition import Oscillator_CalibrationSolution
from ansys.solutions.oscillator_calibration.ui.utils.websockets_helpers import extract_alert_from_alerts_event
from ansys.solutions.oscillator_calibration.utilities.common_functions import LOG_MESSAGE_COLORS, milliseconds

TOAST_STYLE = {"position": "fixed", "top": 130, "right": 10, "width": 350, "zIndex": 1000}

layout = html.Div(
    [
        html.Div(id="alerts_container", style=TOAST_STYLE),
        html.Div(id="trigger_creation_of_alerts_event_listener", style={"display": "none"}),
        html.Div(id="alerts_listener"),
    ]
)


@callback(
    Input("trigger_creation_of_alerts_event_listener", "children"),
    State("url", "pathname"),
    Output("alerts_listener", "children"),
)
def setup_alerts_event_listener(trigger, pathname):
    """Receive alerts from the monitoring step."""
    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    ws = DashClient.create_event_listener(
        project.steps.monitoring_step,
        stream_name=WebsocketStreamNames.ALERTS.value,
        id=WebsocketStreamListeners.ALERTS.value,
    )
    return ws


@callback(
    Output("alerts_container", "children"),
    Input(WebsocketStreamListeners.ALERTS.value, "message"),
)
def update_alerts_view(alerts_event):
    """Update alerts view."""

    alert = extract_alert_from_alerts_event(alerts_event)
    if not alert:
        return no_update

    if time.time() - alert.time > 10:
        # Don't show alerts from more than 10 seconds ago.
        return no_update

    formatted_time = datetime.datetime.fromtimestamp(alert.time).strftime("%Y-%m-%d %H:%M:%S")

    content = html.Div(
        [alert.message, html.Hr(), html.Footer([f"{formatted_time}: {alert.origin}"], style={"font-size": "small"})]
    )

    toast = dbc.Toast(
        content,
        header=alert.level.name.capitalize(),
        is_open=True,
        dismissable=True,
        icon=LOG_MESSAGE_COLORS[alert.level],
        duration=milliseconds(seconds=20),
    )

    return toast
