# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the design table view."""

from ansys.saf.glow.client.dashclient import DashClient, callback
from dash_extensions.enrich import Input, Output, State, callback_context, dash_table, dcc, html, no_update
import pandas as pd

from ansys.solutions.oscillator_calibration.datamodel.shared.websocket_streams import WebsocketStreamListeners
from ansys.solutions.oscillator_calibration.solution.definition import Oscillator_CalibrationSolution
from ansys.solutions.oscillator_calibration.ui.utils.colors import ColorHex, ColorRgba
from ansys.solutions.oscillator_calibration.ui.utils.websockets_helpers import (
    prevent_actor_data_event_update_if_actor_not_updated,
    prevent_event_triggered_update_if_auto_update_is_disabled,
)
from ansys.solutions.oscillator_calibration.utilities.common_functions import (
    get_actor_data,
    get_design_table_data,
    sorted_nicely,
)


def sort_designs(designs: list) -> list:
    sorted_designs = sorted_nicely(designs)
    return [designs.index(design) for design in sorted_designs]


def transform_data(data: dict) -> pd.DataFrame:
    data = pd.DataFrame(data)
    if len(data["Design"]):
        # OptiSLang returns the designs in an unsorted way.
        # The following method organizes the rows in ascending order.
        data = data.loc[sort_designs(data["Design"].to_list())]
        for col, dtype in data.dtypes.items():
            if dtype == "bool":
                data[col] = data[col].astype("str")
            elif dtype == "object":
                data[col] = data[col].astype("str")
        data = data.round(6)
    return data


def design_sort_key(x):
    return tuple(map(int, x.split(".")))


def conditional_design_sort_key(x):
    # Apply the design_sort_key function if column is "Design"
    if x.name == "Design":
        return x.map(design_sort_key)
    else:
        return x


def layout() -> html.Div:
    """Layout of the design table view."""

    style_cell = {
        "textAlign": "left",
        "font_family": "Segoe UI",
        "font_size": "14px",
        "overflowX": "scroll",
        "minWidth": "120px",
        "width": "120px",
        "maxWidth": "120px",
    }

    style_data_conditional = [
        {
            "if": {"column_id": "Design"},
            "minWidth": "80px",
            "width": "80px",
            "maxWidth": "80px",
        },
        {
            "if": {"column_id": "Status", "filter_query": '{Status} eq "Succeeded"'},
            "backgroundColor": "rgb(223, 240, 208)",
            "color": "rgb(0, 0, 0)",
            "minWidth": "100px",
            "width": "100px",
            "maxWidth": "100px",
        },
        {
            "if": {"column_id": "Status", "filter_query": '{Status} eq "Not succeeded"'},
            "backgroundColor": "rgb(254, 221, 215)",
            "color": "rgb(0, 0, 0)",
            "minWidth": "100px",
            "width": "100px",
            "maxWidth": "100px",
        },
        {
            "if": {"column_id": "Feasible", "filter_query": '{Feasible} eq "True"'},
            "backgroundColor": "rgb(223, 240, 208)",
            "color": "rgb(0, 0, 0)",
            "minWidth": "80px",
            "width": "80px",
            "maxWidth": "80px",
        },
        {
            "if": {"column_id": "Feasible", "filter_query": '{Feasible} eq "False"'},
            "backgroundColor": "rgb(254, 221, 215)",
            "color": "rgb(0, 0, 0)",
            "minWidth": "80px",
            "width": "80px",
            "maxWidth": "80px",
        },
    ]

    style_header = {
        "textAlign": "left",
        "font_family": "Segoe UI",
        "font_size": "14px",
        "fontWeight": "bold",
        "whiteSpace": "normal",
        "overflow": "hidden",
        "textOverflow": "ellipsis",
    }

    style_table = {
        "overflowY": "auto",
        "overflowX": "auto",
    }

    table = (
        dash_table.DataTable(
            editable=False,
            id="design_table",
            columns=[],  # set columns dynamically
            data=[],
            fixed_columns={"data": 0},
            fixed_rows={"data": 0, "headers": True},  # Headers stay visible when scrolling
            page_action="none",  # No pagination
            style_as_list_view=True,
            style_cell=style_cell,
            style_data_conditional=style_data_conditional,
            style_header=style_header,
            style_table=style_table,
            virtualization=True,
            sort_action="custom",
            sort_mode="single",
        ),
    )

    design_table_information_loading_div = html.Div(
        id="design_table_information_loading_div", style={"display": "none"}
    )
    design_table_trigger_initial_data_load_div = html.Div(
        id="design_table_trigger_initial_data_load_div", style={"display": "none"}
    )

    design_table_information_loading_spinner = dcc.Loading(
        type="circle",
        fullscreen=True,
        color=ColorHex.ANSYS_ORANGE.value,
        style={
            "background-color": ColorRgba.TRANSPARENT_GREY.value,
        },
        children=[
            design_table_information_loading_div,
            design_table_trigger_initial_data_load_div,
        ],
        delay_show=500,
    )

    return html.Div(
        [
            design_table_information_loading_spinner,
            html.Br(),
            html.Div(table),
        ]
    )


@callback(
    Output("design_table", "data"),
    Output("design_table", "columns"),
    Output("design_table_trigger_initial_data_load_div", "children"),
    Input("selected_state_dropdown", "value"),
    State("url", "pathname"),
    State("navigation_tree", "treeItemClicked"),
)
def initialize_view(actor_hid, pathname, tree_item):
    """Update design table."""

    selected_actor = tree_item["id"]

    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    data, columns = get_design_table_data_and_columns(actor_hid, project.steps.monitoring_step, selected_actor)

    return data.to_dict("records"), columns, "update_required"


@callback(
    Output("design_table_information_loading_div", "children"),
    Input("design_table_trigger_initial_data_load_div", "children"),
    State("selected_state_dropdown", "value"),
    State("url", "pathname"),
    State("navigation_tree", "treeItemClicked"),
    prevent_initial_call=True,
)
def load_information(trigger, actor_hid, pathname, tree_item):
    """Initial information loading for design table."""

    selected_actor = tree_item["id"]
    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    monitoring_step = project.steps.monitoring_step

    if monitoring_step.server_healthy:
        actor_data_files = monitoring_step.actor_data.list_files()
        actor_data = get_actor_data(actor_data_files, selected_actor)

        if len(actor_data.get("states_ids", [])) <= 1:
            # This actor has no or just one, skip
            return no_update

        monitoring_step.actor_to_update = selected_actor
        monitoring_step.hid_to_update = actor_hid
        monitoring_step.update_specific_actor_data()

    return no_update


@callback(
    Output("design_table", "data"),
    Output("design_table", "columns"),
    Input("design_table", "sort_by"),
    Input(WebsocketStreamListeners.ACTOR_DATA.value, "message"),
    State("selected_state_dropdown", "value"),
    State("url", "pathname"),
    State("navigation_tree", "treeItemClicked"),
    State("activate_auto_update", "on"),
    prevent_initial_call=True,
)
def update_view(sort_by, actor_data_event, actor_hid, pathname, tree_item, auto_update):
    """Update design table."""

    prevent_event_triggered_update_if_auto_update_is_disabled(callback_context, auto_update, actor_data_event)

    selected_actor = tree_item["id"]
    prevent_actor_data_event_update_if_actor_not_updated(actor_data_event, callback_context, selected_actor)

    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    monitoring_step = project.steps.monitoring_step
    data, columns = get_design_table_data_and_columns(actor_hid, monitoring_step, selected_actor)

    if sort_by is None or len(sort_by) == 0:
        return data.to_dict("records"), columns

    col_name = sort_by[0]["column_id"]
    direction = sort_by[0]["direction"]
    ascending = True if direction == "asc" else False

    if col_name == "Design":
        df_sorted = data.sort_values(by=col_name, key=lambda x: x.map(design_sort_key), ascending=ascending)
    else:
        df_sorted = data.sort_values(
            by=[col_name, "Design"], ascending=[ascending, ascending], key=conditional_design_sort_key
        )

    return df_sorted.to_dict("records"), columns


def get_design_table_data_and_columns(actor_hid, monitoring_step, selected_actor):
    design_table_data = get_design_table_data(monitoring_step.actor_data.list_files(), selected_actor, actor_hid)

    data = transform_data(design_table_data)
    columns = [{"name": i, "id": i} for i in data.columns]

    return data, columns
