# ©2024, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

from typing import List

from ansys.saf.glow.client.dashclient import DashClient, callback
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
from dash_extensions.enrich import Input, Output, State, callback_context, dcc, html, no_update

from ansys.solutions.oscillator_calibration.datamodel.shared.websocket_streams import WebsocketStreamListeners
from ansys.solutions.oscillator_calibration.solution.definition import Oscillator_CalibrationSolution
from ansys.solutions.oscillator_calibration.ui.utils.websockets_helpers import (
    prevent_actor_data_event_update_if_actor_not_updated,
    prevent_event_triggered_update_if_auto_update_is_disabled,
)
from ansys.solutions.oscillator_calibration.utilities.common_functions import get_actor_data


def get_default_dropdown_style():
    return {
        "textAlign": "left",
        "width": "200px",
        "margin-right": "30px",
        "height": "40px",
    }


def layout(selected_actor: str, project: Oscillator_CalibrationSolution) -> html.Div:
    annotated_options = get_annotated_options(selected_actor, project)
    selected_hid = determine_selected_hid_for_fresh_page(annotated_options)
    return html.Div(
        [
            html.Hr(className="my-2"),
            dbc.Stack(
                id="monitoring_page_footer_buttons",
                children=dcc.Dropdown(
                    options=annotated_options,
                    value=selected_hid,
                    id="selected_state_dropdown",
                    disabled=False,
                    clearable=False,
                    searchable=True,
                    style=get_default_dropdown_style(),
                    placeholder="Select state",
                ),
                direction="horizontal",
                gap=1,
            ),
        ],
        style={"padding-right": "2%"},
    )


@callback(
    Output("selected_state_dropdown", "options"),
    Output("selected_state_dropdown", "value"),
    Input(WebsocketStreamListeners.ACTOR_DATA.value, "message"),
    State("navigation_tree", "treeItemClicked"),
    State("selected_state_dropdown", "value"),
    State("url", "pathname"),
    State("activate_auto_update", "on"),
)
def display_state_selection_drop_down(actor_data_event, tree_item, current_selected_state, pathname, auto_update):
    """Populate and display state selection drop down."""

    triggered_id = callback_context.triggered_id

    if not triggered_id:
        raise PreventUpdate

    prevent_event_triggered_update_if_auto_update_is_disabled(callback_context, auto_update)

    actor_id = tree_item["id"]
    prevent_actor_data_event_update_if_actor_not_updated(actor_data_event, callback_context, actor_id)

    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    annotated_options = get_annotated_options(actor_id, project)
    selected_hid = determine_selected_hid_for_page_update(annotated_options, current_selected_state)

    return (annotated_options, selected_hid)


def get_annotated_options(selected_actor: str, project: Oscillator_CalibrationSolution) -> List[dict[str, str]]:

    monitoring_step = project.steps.monitoring_step
    actor_data = get_actor_data(monitoring_step.actor_data.list_files(), selected_actor)

    id_options = []
    for state_id in actor_data.get("states_ids", []):
        sub_hids = state_id.split(".")
        sorting_index = ()
        for sub_hid in sub_hids:
            sorting_index = sorting_index + (int(sub_hid),)

        id_options.append(
            {
                "sorting_index": sorting_index,
                "state_id": state_id,
                "label": f"State: {state_id}",
            }
        )

    annotated_options = []
    if id_options:
        sorted_options = sorted(id_options, key=lambda state_entry: state_entry["sorting_index"])
        annotated_options = [
            {
                "label": state["label"],
                "value": state["state_id"],
            }
            for state in sorted_options
        ]

    return annotated_options


def determine_selected_hid_for_fresh_page(annotated_options: List[dict[str, str]]) -> str:

    selected_hid = None
    if annotated_options:
        selected_hid = annotated_options[0]["value"]

    return selected_hid


def determine_selected_hid_for_page_update(annotated_options: List[dict[str, str]], current_selected_state: str) -> str:

    selected_hid = no_update
    if current_selected_state is None and annotated_options:
        selected_hid = annotated_options[0]["value"]
    elif current_selected_state is not None and not annotated_options:
        selected_hid = None

    return selected_hid
