# ©2024, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import webbrowser

from ansys.saf.glow.client.dashclient import DashClient, callback
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
from dash_extensions.enrich import Input, Output, State, html

from ansys.solutions.oscillator_calibration.solution.definition import Oscillator_CalibrationSolution
from ansys.solutions.oscillator_calibration.ui.components.logs_table import LogsTable
from ansys.solutions.oscillator_calibration.utilities.common_functions import try_read_text


def get_default_button_style():
    return {
        "display": "flex",
        "justify-content": "center",
        "align-items": "center",
        "color": "rgba(0, 0, 0, 1)",
        "background-color": "rgba(255, 255, 255, 1)",
        "border-color": "rgba(0, 0, 0, 1)",
        "height": "40px",
    }


def layout(portal_ui_url: str) -> html.Div:
    optislang_logs_button = dbc.Button(
        "optiSLang logs",
        id="osl_logs_button",
        n_clicks=0,
        style=get_default_button_style(),
        size="sm",
    )

    open_in_browser_button = dbc.Button(
        "Open in browser", id="open_in_browser", style=get_default_button_style(), n_clicks=0, size="sm"
    )
    if not portal_ui_url:
        open_in_browser_button.style["display"] = "none"

    page_footer_buttons = dbc.Stack(
        id="page_footer_buttons",
        children=[
            optislang_logs_button,
            open_in_browser_button,
        ],
        direction="horizontal",
        gap=1,
    )

    osl_logs_collapse = dbc.Collapse(
        id="osl_logs_collapse",
        children=[],
        is_open=False,
    )
    divider = html.Hr(className="my-2")

    return html.Div(
        [
            divider,
            page_footer_buttons,
            osl_logs_collapse,
        ],
        style={"padding-right": "2%"},
    )


@callback(
    Output("osl_logs_collapse", "children"),
    Output("osl_logs_collapse", "is_open"),
    Input("osl_logs_button", "n_clicks"),
    Input("url", "pathname"),
    State("osl_logs_collapse", "is_open"),
    prevent_initial_call=True,
)
def display_optislang_logs(n_clicks, pathname, is_open):
    """Display optiSLang logs."""
    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step

    if not n_clicks:
        # Button has never been clicked
        return None, False

    if problem_setup_step.project_locked:
        problem_setup_step.fetch_optislang_logs()
        osl_logs = try_read_text(problem_setup_step.osl_log_file).splitlines()
        osl_logs = [log.strip() for log in osl_logs if log]
        table = LogsTable(osl_logs)
        return table.render(), not is_open

    raise PreventUpdate


@callback(
    Output("open_in_browser", "children"),
    Input("open_in_browser", "n_clicks"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def open_in_browser(n_clicks, pathname):
    """Open the Portal UI in browser view."""
    portal_ui_url = DashClient.get_portal_ui_url()
    webbrowser.open_new(portal_ui_url)

    raise PreventUpdate
