# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the problem setup step."""

from copy import deepcopy
import os
from typing import Any, Dict, List, Optional, Tuple

from ansys.saf.glow.client.dashclient import DashClient, callback
from ansys.solutions.optislang.frontend_components.load_sections import to_dash_sections, update_designs_to_dash_section
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
from dash_extensions.enrich import ALL, MATCH, Input, Output, State, callback_context, html, no_update

from ansys.solutions.oscillator_calibration.solution.definition import Oscillator_CalibrationSolution
from ansys.solutions.oscillator_calibration.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.oscillator_calibration.solution.step_slices import PlaceholderSlice


def layout(problem_setup_step: ProblemSetupStep) -> html.Div:
    """Layout of the problem setup step."""
    placeholder_slice = PlaceholderSlice.create(problem_setup_step)
    
    if placeholder_slice.ui_placeholders and not placeholder_slice.project_locked:
        problem_setup_step.update_osl_placeholders_with_ui_values()
        placeholder_slice = PlaceholderSlice.create(problem_setup_step)

    project_properties_sections = to_dash_sections(
        placeholder_slice.placeholders, placeholder_slice.registered_files, placeholder_slice.project_locked
    )

    return html.Div(
        [
            # Input form
            html.P("Input Form", style={"font-size": "18px"}),
            html.Hr(className="my-2"),
            dbc.Row(id="osl-dash-sections", children=project_properties_sections),
            dbc.Button(
                id="dummy_button_2",
                style={"display": "none"},
                n_clicks=0,
            ),
        ]
    )


def get_row_and_col_id(full_id: str) -> Tuple[str, str]:
    split_values = full_id.split("##")
    values = split_values[1:]
    rowId = values[0].split("#")[1]
    colId = values[1].split("#")[1]
    return rowId, colId


def get_parameter_manager_and_key(full_id: str) -> Tuple[str, str]:
    split_values = full_id.split("##")
    values = split_values[1:]
    pm_name = values[0].split("#")[0]
    key = values[0].split("#")[1]
    return pm_name, key


def get_start_designs_values_dict(
    dash_values: List[Any], full_ids: List[Dict[str, str]], order: Optional[List[str]] = None
) -> Dict[str, Dict[str, Any]]:
    ids = [id["placeholder"] for id in full_ids]

    starts_design_values = {
        get_row_and_col_id(id): convert_dash_value(dv)
        for id, dv in zip(ids, dash_values)
        if id.startswith("StartDesigns")
    }

    designs: Dict[str, Dict[str, Any]] = {}

    for (row, col), value in starts_design_values.items():
        if row in designs:
            designs[row][col] = value
        else:
            designs[row] = {col: value}

    if not order:
        row = next(iter(designs), None)
        order = [col for col in designs[row]] if row else []
    ordered_designs: Dict[str, Dict[str, Any]] = {}

    def get_index(x: str):
        if x in order:
            return order.index(x)
        else:
            return float("inf")

    for row in designs:
        ordered_designs[row] = dict(sorted(designs[row].items(), key=lambda x: get_index(x[0])))

    return ordered_designs


def get_placeholder_column_order(placeholders: dict) -> Optional[List[str]]:

    if not "placeholder_values" in placeholders:
        return None
    placeholder_values = placeholders["placeholder_values"]

    if not "StartDesigns" in placeholder_values:
        return None

    start_designs = placeholder_values["StartDesigns"]
    first_start_design = next(iter(start_designs), None)

    if first_start_design:
        order = [q["First"] for q in first_start_design["parameters"]["sequence"]]
        return order
    else:
        return None


def convert_dash_value(value: Any) -> Any:
    """Boolean checkboxes in Dash are a list: [] or [True], we need to convert these to boolean for the backend."""
    if isinstance(value, list) and value:
        value = True
    elif isinstance(value, list) and not value:
        value = False
    return value


@callback(
    Output("start_analysis", "disabled"),
    Input("dummy_button_2", "n_clicks"),
    State({"container": "placeholders", "placeholder": ALL}, "value"),
    State({"container": "placeholders", "placeholder": ALL}, "id"),
    State({"type": "upload", "placeholder": ALL}, "id"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def initialize_dictionary_of_ui_placeholders(n_clicks, data, ids, input_file_ids, pathname):
    """This initializes the dictionary of ui placeholders using only the values in the Placeholders section."""
    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step
    designs = {}
    parameters = {}
    if problem_setup_step.ui_placeholders == {}:
        ui_data = {}
        for index in range(len(data)):
            name = ids[index]["placeholder"]
            dash_value = data[index]
            value = convert_dash_value(dash_value)
            if name.startswith("StartDesigns"):
                rowId, key = get_row_and_col_id(name)
                if rowId in designs:
                    designs[rowId].update({key: value})
                else:
                    designs[rowId] = {key: value}
            elif name.startswith("ParameterManager"):
                placeholder_name, key = get_parameter_manager_and_key(name)
                parameters[key] = value
                ui_data.update({placeholder_name: parameters})
            else:
                ui_data.update({ids[index]["placeholder"]: value})
            ui_data.update({"StartDesigns": designs})
            ui_data["start_analysis_requested"] = False
        
        analysis_locked = bool(input_file_ids)

        for index in range(len(input_file_ids)):
            ui_data.update({input_file_ids[index]["placeholder"]: ""})

        problem_setup_step.analysis_locked = analysis_locked

        problem_setup_step.ui_placeholders = ui_data
        problem_setup_step.update_osl_placeholders_with_ui_values()

        return analysis_locked
    else:
        return no_update


@callback(
    Output({"container": "placeholders", "placeholder": MATCH}, "value"),
    Input({"container": "placeholders", "placeholder": MATCH}, "value"),
    State({"container": "placeholders", "placeholder": MATCH}, "id"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def update_ui_placeholders(dash_value, full_id, pathname):
    """
    This updates the dictionary of ui placeholders each time the ui data changes in the Placeholders section.
    If the value entered is None, "", or invalid, the last value entered in the component is returned.
    """
    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step
    name = full_id["placeholder"]
    ui_data = problem_setup_step.ui_placeholders
    value = convert_dash_value(dash_value)

    if value is None or value == "":
        if name.startswith("StartDesigns"):
            rowId, key = get_row_and_col_id(name)
            last_value = ui_data["StartDesigns"][rowId][key]
        elif name.startswith("ParameterManager"):
            pm_name, key = get_parameter_manager_and_key(name)
            last_value = ui_data[pm_name][key]
        else:
            last_value = ui_data[name]
        return last_value
    else:

        if name.startswith("StartDesigns"):
            rowId, key = get_row_and_col_id(name)
            ui_data["StartDesigns"][rowId][key] = value
        elif name.startswith("ParameterManager"):
            pm_name, key = get_parameter_manager_and_key(name)
            ui_data[pm_name][key] = value
        else:
            ui_data.update({name: value})
        problem_setup_step.ui_placeholders = ui_data
        return no_update


@callback(
    Output({"type": "upload", "placeholder": MATCH}, "children"),
    Input({"type": "upload", "placeholder": MATCH}, "isCompleted"),
    State({"type": "upload", "placeholder": MATCH}, "fileNames"),
    State({"type": "upload", "placeholder": MATCH}, "upload_id"),
    State({"type": "upload", "placeholder": MATCH}, "id"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def upload_input_files_to_project_and_update_ui_placeholders(
    is_completed, filenames, upload_id, component_id, pathname
):
    """This uploads an Input file to the project directory and updates the dictionary of ui placeholders every time
    the ui data changes in the Input files section."""
    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step

    name = component_id["placeholder"]
    if is_completed and filenames and len(filenames):
        filename = filenames[0]
        filepath = os.path.join(problem_setup_step.upload_directory, upload_id, filename)
        file_data = open(filepath, mode="rb")

        project.upload_file(f"Problem_Setup/Input_Files/{filename}", file_data)
        ui_data = problem_setup_step.ui_placeholders
        ui_data.update({name: filename})
        problem_setup_step.ui_placeholders = ui_data

        return no_update
    else:
        raise PreventUpdate


@callback(
    Output("start_analysis", "disabled"),
    Input({"type": "upload", "placeholder": ALL}, "isCompleted"),
    State({"type": "upload", "placeholder": ALL}, "fileNames"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def update_status_of_start_analysis_button(is_completed, filenames, pathname):
    """This enables the start analysis button if a file has been uploaded for all input files."""
    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step
    analysis_locked = not all(filenames)
    problem_setup_step.analysis_locked = analysis_locked

    return analysis_locked


@callback(
    Output("dummy_button_2", "n_clicks"),
    Input("url", "pathname"),
    State("dummy_button_2", "n_clicks"),
)
def on_page_load_initialize_dictionary_of_ui_placeholder_values(pathname, n_clicks):
    return n_clicks + 1


@callback(
    Output("start-designs-view", "children"),
    Input({"type": "button-add", "index": ALL}, "n_clicks"),
    Input({"type": "button-del", "index": ALL}, "n_clicks"),
    State("url", "pathname"),
    State({"container": "placeholders", "placeholder": ALL}, "value"),
    State({"container": "placeholders", "placeholder": ALL}, "id"),
    prevent_initial_call=True,
)
def update_start_designs_table(n_clicks_add, n_clicks_del, pathname, dash_values: List[Any], ids: List[Dict[str, str]]):
    """This updates the StartDesigns table every time the '+' or '-' button is clicked."""
    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step
    slice = PlaceholderSlice.create(problem_setup_step)

    placeholders = slice.placeholders
    ui_data = slice.ui_placeholders

    order = get_placeholder_column_order(placeholders)
    designs = get_start_designs_values_dict(dash_values, ids, order)

    button_type = str(callback_context.triggered_id["type"])
    row_id = str(callback_context.triggered_id["index"])

    if button_type == "button-add":
        designs = add_row(designs, row_id)
    elif button_type == "button-del":
        designs = delete_row(designs, row_id)
    else:
        raise ValueError(f"Unexpected button type {button_type} for id {row_id}")

    ui_data.update({"StartDesigns": designs})

    problem_setup_step.ui_placeholders = ui_data

    problem_setup_step.update_osl_placeholders_with_ui_values()

    post_update_slice = PlaceholderSlice.create(problem_setup_step)

    return update_designs_to_dash_section(post_update_slice.placeholders, post_update_slice.project_locked)


def add_row(designs: Dict[str, Any], id_: str) -> Dict[str, Any]:
    new_value = deepcopy(designs[id_])  # Value from the target key
    new_designs = {}
    for key, value in designs.items():
        new_designs[key] = value
        if key == id_:
            new_designs["temp"] = new_value
    new_designs = renumber_designs(new_designs)
    return new_designs


def delete_row(designs: Dict[str, Any], id_: str) -> Dict[str, Any]:
    del designs[id_]
    designs = renumber_designs(designs)
    return designs


def renumber_designs(designs: dict) -> Dict[str, Any]:
    renumbered_designs = {str(index): value for index, (_, value) in enumerate(designs.items(), start=1)}
    return renumbered_designs
