# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the project result files view."""

import base64
import logging
import os
import time
from typing import List, Optional

from ansys.saf.glow.client.dashclient import DashClient, callback
from ansys.saf.glow.solution import FileReference
import dash_bootstrap_components as dbc
from dash_extensions.enrich import Input, Output, State, callback_context, dash_table, dcc, html, no_update
import pandas as pd
from retrying import retry

from ansys.solutions.oscillator_calibration.datamodel.shared.models import ActorAndHid
from ansys.solutions.oscillator_calibration.datamodel.shared.websocket_streams import WebsocketStreamListeners
from ansys.solutions.oscillator_calibration.solution.definition import Oscillator_CalibrationSolution
from ansys.solutions.oscillator_calibration.solution.result_files_step import FileState, RegisteredFile, ResultFilesStep
from ansys.solutions.oscillator_calibration.ui.components.button_group import ResultfilesButtons
from ansys.solutions.oscillator_calibration.ui.utils.colors import ColorHex, ColorRgba
from ansys.solutions.oscillator_calibration.ui.utils.websockets_helpers import (
    prevent_event_triggered_update_if_auto_update_is_disabled,
)
from ansys.solutions.oscillator_calibration.utilities.deployment import is_containerized

logger = logging.getLogger(__name__)


FileStateIcons = {
    FileState.PENDING.value: "⧖",
    FileState.MISSING.value: "⚠",
    FileState.FINISHED.value: "✓",
}


# Max size, in bytes, for which to allow preview for a given extension
# Too large files risk making the app unstable

MB = int(1e6)
DEFAULT_MAX_PREVIEW_SIZE = 10 * MB

VIDEO_LIMIT = 120 * MB
PICTURE_LIMIT = 40 * MB
JSON_LIMIT = 25 * MB

MAX_PREVIEW_SIZE_BY_EXTENSION = {
    "json": JSON_LIMIT,
    # Videos
    "mp4": VIDEO_LIMIT,
    "mpeg": VIDEO_LIMIT,
    "ogv": VIDEO_LIMIT,
    "webm": VIDEO_LIMIT,
    # Pictures
    "jpeg": PICTURE_LIMIT,
    "jpg": PICTURE_LIMIT,
    "png": PICTURE_LIMIT,
    "svg": PICTURE_LIMIT,
    "gif": PICTURE_LIMIT,
    "bmp": PICTURE_LIMIT,
    "ico": PICTURE_LIMIT,
    "apng": PICTURE_LIMIT,
    "avif": PICTURE_LIMIT,
    "webp": PICTURE_LIMIT,
    "tif": PICTURE_LIMIT,
    "tiff": PICTURE_LIMIT,
}


def _file_too_large_to_preview(file_extension: str, file_size_bytes: int) -> bool:
    max_size = DEFAULT_MAX_PREVIEW_SIZE
    if file_extension in MAX_PREVIEW_SIZE_BY_EXTENSION:
        max_size = MAX_PREVIEW_SIZE_BY_EXTENSION[file_extension]

    return file_size_bytes > max_size


def _format_size(size_bytes: int, suffix="B") -> str:
    """Scaling bytes to its proper format e.g. 1253656 => "1.2MB", 125365 => "125.4KB" """
    factor = 1024
    for unit in ["", "K", "M", "G", "T", "P"]:
        if size_bytes < factor:
            return f"{size_bytes:.1f}{unit}{suffix}"
        size_bytes /= factor


style_cell = {
    "textAlign": "left",
    "font_family": "Segoe UI",
    "font_size": "14px",
    "border-right": "none",
}

style_data_conditional = [
    {
        "if": {
            "column_id": "copy_link",
        },
        "textAlign": "center",
        "userSelect": "none",
        "cursor": "pointer",
    }
]

style_header = {
    "textAlign": "left",
    "font_family": "Segoe UI",
    "font_size": "15px",
    "fontWeight": "bold",
    "border-right": "none",
}

hidden_element_style = {"display": "none"}

result_files_table = dcc.Loading(
    type="circle",
    color=ColorHex.ANSYS_ORANGE.value,
    children=[
        dash_table.DataTable(
            id="datatable-interactivity",
            columns=[],
            data=[],
            row_selectable="single",
            selected_rows=[],
            page_action="native",
            page_current=0,
            page_size=10,
            style_as_list_view=True,
            style_header=style_header,
            style_cell=style_cell,
        )
    ],
)

extra_message_selected_file = html.Div(
    id="selected-file-message",
    style={"min-height": "1.5rem"},  # Set min-height too avoid content jumping
)

table_loading_message = html.Div(
    id="table-loading-message",
    style={"min-height": "1.5rem"},  # Set min-height too avoid content jumping
)

buttons = ResultfilesButtons.create_buttons()

modal = dbc.Modal(
    [
        dbc.ModalHeader(id="preview-header"),
        dbc.ModalBody(id="preview-body"),
        dbc.ModalFooter(dbc.Button("Close", id="close-preview-btn", className="ml-auto", color="secondary")),
    ],
    id="preview-modal",
    is_open=False,
    scrollable=True,
    size="xl",
)

clipboard_target = html.Div(id="clipboard-target", style=hidden_element_style)
download_div = dcc.Download(id="download-text")

spinner = dcc.Loading(
    type="circle",
    fullscreen=True,
    color=ColorHex.ANSYS_ORANGE.value,
    style={
        "background-color": ColorRgba.TRANSPARENT_GREY.value,
    },
    children=[modal, download_div, clipboard_target],
)


hidden_divs = html.Div(id="hidden-div", style=hidden_element_style)
checkbox_states = html.Div(id="checkbox-states", style=hidden_element_style)

layout = html.Div(
    [
        extra_message_selected_file,
        buttons,
        result_files_table,
        table_loading_message,
        spinner,
        hidden_divs,
        checkbox_states,
    ]
)


@callback(
    Output("download-btn", "disabled"),
    Output("preview-btn", "disabled"),
    Output("copy-link-btn", "disabled"),
    Output("selected-file-message", "children"),
    Input("datatable-interactivity", "selected_rows"),
    State("datatable-interactivity", "data"),
)
def update_button_status(selected_rows, rows):
    if len(selected_rows) != 1:
        return True, True, True, ""

    # Copy link does not work on containerized, need to disable
    disable_copy_button = is_containerized()

    if selected_rows:
        message = ""
        selected_row_data = rows[selected_rows[0]]
        file_path = selected_row_data["Filename"]
        file_extension = file_path.split(".")[-1].lower()
        is_previewable = file_extension in [
            "json",
            "jpg",
            "png",
            "jpeg",
            "txt",
            "log",
            "csv",
            "xml",
            "apng",
            "avif",
            "gif",
            "svg",
            "webp",
            "bmp",
            "ico",
            "tif",
            "tiff",
            "mp4",
            "mpeg",
            "ogv",
            "webm",
        ]

        too_large = _file_too_large_to_preview(file_extension, selected_row_data["size"])

        if too_large:
            message = "The file is too large to preview. You may download it."
            is_previewable = False

        disable_preview_button = not is_previewable

        if selected_row_data["status"] == FileState.FINISHED.value:
            return False, disable_preview_button, disable_copy_button, message
        else:
            return (
                True,
                True,
                True,
                message,
            )


def make_downloadable(result_files: ResultFilesStep, file: RegisteredFile):
    file_tag = file["tag"]
    result_files.file_to_download = file_tag
    result_files.download()
    time.sleep(1.0)  # Sometimes need to wait for a bit for the file to be visible


@retry(stop_max_attempt_number=4)
def is_downloadable(result_files: ResultFilesStep, file: RegisteredFile) -> bool:
    file_name = file["Filename"]

    files = result_files.downloads.list_files()

    is_downloadable = any(file_name in f.url for f in files)
    return is_downloadable


@retry(stop_max_attempt_number=4)
def get_downloadable_file_reference(result_files: ResultFilesStep, file: RegisteredFile) -> Optional[FileReference]:
    file_name = file["Filename"]

    files = result_files.downloads.list_files()

    file_reference = next((f for f in files if file_name in f.url), None)
    return file_reference


def download_file_content(result_files: ResultFilesStep, file: RegisteredFile) -> bytes:
    make_downloadable(result_files, file)
    file_reference = get_downloadable_file_reference(result_files, file)
    file_content = file_reference.read_bytes()

    return file_content


@callback(
    Output("preview-modal", "is_open"),
    Output("preview-body", "children"),
    Output("preview-header", "children"),
    Input("preview-btn", "n_clicks"),
    Input("close-preview-btn", "n_clicks"),
    State("datatable-interactivity", "selected_rows"),
    State("datatable-interactivity", "data"),
    State("preview-modal", "is_open"),
    State("url", "pathname"),
)
def toggle_preview_modal(
    n_clicks, close_clicks, selected_rows: List[int], rows_data: List[RegisteredFile], is_open, pathname
):
    if n_clicks and selected_rows:
        selected_file = rows_data[selected_rows[0]]
        file_name = selected_file["Filename"]

        result_files = DashClient[Oscillator_CalibrationSolution].get_project(pathname).steps.result_files_step

        file_extension = os.path.splitext(file_name)[1].lower().lstrip(".")

        try:
            file_content = download_file_content(result_files, selected_file)
            if file_extension in [
                "jpeg",
                "jpg",
                "png",
                "svg",
                "gif",
                "bmp",
                "ico",
                "apng",
                "avif",
                "webp",
                "tif",
                "tiff",
            ]:
                mime_type = "image/svg+xml" if file_extension == "svg" else f"image/{file_extension}"
                b64_encoded_image = base64.b64encode(file_content)
                decoded = b64_encoded_image.decode("ascii")
                img_src = f"data:{mime_type};base64,{decoded}"
                preview_content = html.Img(src=img_src, style={"max-width": "100%", "height": "auto"})
            elif file_extension in ["mp4", "mpeg", "ogv", "webm"]:
                b64_encoded_video = base64.b64encode(file_content)
                decoded = b64_encoded_video.decode("ascii")
                video_src = f"data:video/{file_extension};base64,{decoded}"
                preview_content = html.Video(
                    src=video_src, controls=True, style={"max-width": "100%", "height": "auto"}
                )
            elif file_extension == "json":
                preview_content = file_content.decode("utf-8")
                preview_content = html.Pre(preview_content, style={"white-space": "pre-wrap", "overflow-x": "auto"})
            else:
                try:
                    decoded = file_content.decode("ascii")
                    preview_content = html.Pre(decoded, style={"white-space": "pre-wrap", "word-wrap": "break-word"})
                except:
                    preview_content = f"Unable to preview file {file_name}"

        except Exception as e:
            logger.warn(e)
            preview_content = f"Error loading file {file_name}"

        return not is_open, preview_content, file_name

    elif close_clicks:
        return False, None, None
    return is_open, None, None


@callback(
    Output("download-text", "data"),
    Input("download-btn", "n_clicks"),
    State("datatable-interactivity", "selected_rows"),
    State("datatable-interactivity", "data"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def download_file(n_clicks, selected_rows: List[int], table_data: List[RegisteredFile], pathname):
    if selected_rows is None or not selected_rows:
        return None

    result_files = DashClient[Oscillator_CalibrationSolution].get_project(pathname).steps.result_files_step

    selected_row_index = selected_rows[0]

    try:
        file = table_data[selected_row_index]
        file_content = download_file_content(result_files, file)
        return dcc.send_bytes(file_content, file["Filename"])
    except Exception as e:
        logger.error(f"Error downloading file: {e}")
        return None


@callback(
    Output("clipboard-target", "children"),
    Input("copy-link-btn", "n_clicks"),
    State("datatable-interactivity", "selected_rows"),
    State("datatable-interactivity", "data"),
    State("url", "pathname"),
)
def update_link_to_copy(n_clicks, selected_rows: List[int], table_data: List[RegisteredFile], pathname):
    if not selected_rows:
        return

    result_files = DashClient[Oscillator_CalibrationSolution].get_project(pathname).steps.result_files_step
    selected_row_index = selected_rows[0]
    file = table_data[selected_row_index]

    # Check if file is already available
    downloadable = is_downloadable(result_files, file)

    if not downloadable:
        # File is not in downloads folder, so make it available
        make_downloadable(result_files, file)

    file_reference = get_downloadable_file_reference(result_files, file)
    pd_copy_link = pd.DataFrame([file_reference.url])
    pd_copy_link.to_clipboard(index=False, header=False)
    return no_update


@callback(
    Output("datatable-interactivity", "data"),
    Output("datatable-interactivity", "columns"),
    Output("table-loading-message", "children"),
    Input(WebsocketStreamListeners.PROJECT_STATE.value, "message"),
    Input("selected_state_dropdown", "value"),
    State("url", "pathname"),
    State("navigation_tree", "treeItemClicked"),
    State("monitoring_tabs", "active_tab"),
    State("activate_auto_update", "on"),
)
def update_data_table(project_state_event, selected_hid, pathname, treeItem, selected_tab, auto_update):
    """Update project result file table."""
    prevent_event_triggered_update_if_auto_update_is_disabled(callback_context, auto_update)

    result_files = DashClient[Oscillator_CalibrationSolution].get_project(pathname).steps.result_files_step

    message = ""

    if selected_tab == "project_result_files_tab":
        result_files.populate_project_file_details()
        registered_files = result_files.registered_files
    else:
        if selected_hid is not None:
            selected_actor = treeItem["id"]
            selected_actor_and_hid = ActorAndHid(actor_uid=selected_actor, hid=selected_hid)
            result_files.selected_actor_and_hid = selected_actor_and_hid

            try:
                result_files.populate_actor_file_details()
                registered_files = result_files.registered_files
            except:
                message = "An error occurred loading the registered files"
                registered_files = []
        else:
            # For a few cases, hid is None, and we cannot get the working directory
            registered_files = []

    if not registered_files:
        updated_data = pd.DataFrame(columns=[])
        columns = []
        if not message:
            # Write a message if no error message is present
            message = "No files to show for this actor"
    else:
        updated_data = pd.DataFrame(registered_files)

        status_icon = updated_data["status"].apply(lambda status: FileStateIcons[status])
        updated_data.insert(2, "Status", status_icon)

        formatted_size = updated_data["size"].apply(lambda size: _format_size(size))
        updated_data.insert(4, "Size", formatted_size)
        columns = [
            {"name": i, "id": i}
            for i in updated_data.columns
            if i not in ["location", "status", "basepathmode", "tag", "size"]
        ]

    data = updated_data.to_dict("records")

    return data, columns, message
