# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the node status overview view."""

import logging

from ansys.saf.glow.client.dashclient import DashClient, callback
from dash.exceptions import PreventUpdate
from dash_extensions.enrich import Input, Output, State, callback_context, dcc, html
import optislang_dash_lib

from ansys.solutions.oscillator_calibration.datamodel.shared.models import ActorAndHid
from ansys.solutions.oscillator_calibration.datamodel.shared.websocket_streams import (
    ServerHealthyEvent,
    WebsocketStreamListeners,
)
from ansys.solutions.oscillator_calibration.solution.command_step import execute_command, get_startable_actors_and_hids
from ansys.solutions.oscillator_calibration.solution.definition import Oscillator_CalibrationSolution
from ansys.solutions.oscillator_calibration.ui.utils.colors import ColorHex, ColorRgba
from ansys.solutions.oscillator_calibration.ui.utils.websockets_helpers import (
    get_value_from_websocket_event_or_step,
    prevent_event_triggered_update_if_auto_update_is_disabled,
)
from ansys.solutions.oscillator_calibration.utilities.common_functions import (
    find_node_by_uid,
    get_root_node_from_full_project_status_info,
    try_deserialize,
)

COMMANDS = {
    "START": "start",
    "STOP": "stop",
    "RESET": "reset",
    "SETSUCCEEDED": "set_succeeded_state",
}
START_COMMAND = "START"
ACTOR_UID_KEY = "actor_uid"
HID_KEY = "hid"
COMMAND_KEY = "command"
ELEMENTS_KEY = "elements"

logger = logging.getLogger(__name__)


def layout() -> html.Div:
    """Layout of the status overview view."""

    empty_project_status_info = {}

    node_status_view = optislang_dash_lib.Nodestatusviewcomponent(
        id="node-status-view-component",
        project_state=empty_project_status_info,
        server_interaction_possible=True,
    )

    command_execution_spinner = dcc.Loading(
        type="circle",
        fullscreen=True,
        color=ColorHex.ANSYS_ORANGE.value,
        style={
            "background-color": ColorRgba.TRANSPARENT_GREY.value,
        },
        children=[
            html.Div(id="command_execution_done", style={"display": "none"}),
        ],
    )

    node_status_div = html.Div(
        id="status-overview-component-div1",
        children=[node_status_view, command_execution_spinner],
        style={"display": "block"},
    )

    return node_status_div


@callback(
    Output("node-status-view-component", "project_state"),
    Input(WebsocketStreamListeners.FULL_PROJECT_INFORMATION.value, "message"),
    State("url", "pathname"),
    State("navigation_tree", "treeItemClicked"),
    State("activate_auto_update", "on"),
)
def update_node_status_view(full_project_status_information_updated, pathname, tree_item, auto_update):
    """Update Node status component."""
    prevent_event_triggered_update_if_auto_update_is_disabled(callback_context, auto_update)

    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    monitoring_step = project.steps.monitoring_step

    actor_uid = tree_item["id"]
    full_project_status_info = try_deserialize(monitoring_step.full_project_status_information)
    project_status_info = get_node_status_info_for_display(actor_uid, full_project_status_info)

    return project_status_info


@callback(
    Output("node-status-view-component", "server_interaction_possible"),
    Input(WebsocketStreamListeners.SERVER_HEALTH.value, "message"),
    State("url", "pathname"),
)
def update_server_interaction_possible(server_healthy_event, pathname):
    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    server_healthy = get_value_from_websocket_event_or_step(
        server_healthy_event, project.steps.monitoring_step, ServerHealthyEvent
    )

    return server_healthy


@callback(
    Input("node-status-view-component", "server_command_trigger"),
    Output("command_execution_done", "children"),
    State("url", "pathname"),
)
def execute_server_command(server_command_trigger, pathname):

    if server_command_trigger is None:
        raise PreventUpdate

    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)

    command = server_command_trigger[COMMAND_KEY]
    elements = server_command_trigger[ELEMENTS_KEY]

    osl_command = COMMANDS.get(command)
    if not osl_command:
        logger.error(f"Requested server command {command} is not supported. Command is not executed.")
        return

    command_step = project.steps.command_step
    actors_and_hids = []
    for element in elements:
        actor = element[ACTOR_UID_KEY]
        hid = element[HID_KEY]
        actors_and_hids.append(ActorAndHid(actor_uid=actor, hid=hid))

    if command == START_COMMAND:
        startable_actors_and_hids = get_startable_actors_and_hids(command_step, actors_and_hids)

        if startable_actors_and_hids is None:
            logger.warn(f"Could not get list of startable actors")
            return

        for actor_and_hid in actors_and_hids:
            if actor_and_hid not in startable_actors_and_hids:
                logger.warn(f'Command "START" selected for {actor_and_hid} but is not supported for this actor/hid.')

        if len(startable_actors_and_hids) == 0:
            logger.warn(f"No startable actors selected")
            return

        actors_and_hids = startable_actors_and_hids

    execute_command(command_step, osl_command, actors_and_hids)

    return server_command_trigger


def get_node_status_info_for_display(current_actor_uid: str, full_project_status_info: dict) -> dict:

    try:
        root_node = get_root_node_from_full_project_status_info(full_project_status_info)
    except:
        raise PreventUpdate

    current_node_status_info = find_node_by_uid(root_node, current_actor_uid)
    if not current_node_status_info:
        raise PreventUpdate

    return current_node_status_info
