# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the summary view."""
from typing import Tuple

from ansys.saf.glow.client.dashclient import DashClient, callback
from ansys.saf.glow.solution import FileReference
import dash_bootstrap_components as dbc
from dash_extensions.enrich import Input, Output, State, callback_context, dcc, html, no_update
import pandas as pd

from ansys.solutions.oscillator_calibration.datamodel import datamodel
from ansys.solutions.oscillator_calibration.datamodel.shared.websocket_streams import WebsocketStreamListeners
from ansys.solutions.oscillator_calibration.solution.definition import Oscillator_CalibrationSolution
from ansys.solutions.oscillator_calibration.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.oscillator_calibration.ui.components.actor_information_table import ActorInformationTableAIO
from ansys.solutions.oscillator_calibration.ui.components.actor_logs_table import ActorLogsTableAIO
from ansys.solutions.oscillator_calibration.ui.components.actor_statistics_table import ActorStatisticsTableAIO
from ansys.solutions.oscillator_calibration.ui.components.node_control import layout as node_control_layout
from ansys.solutions.oscillator_calibration.ui.components.system_files_download import layout as system_files_download_layout
from ansys.solutions.oscillator_calibration.ui.utils.colors import ColorHex, ColorRgba
from ansys.solutions.oscillator_calibration.ui.utils.websockets_helpers import (
    prevent_actor_data_event_update_if_actor_not_updated,
    prevent_event_triggered_update_if_auto_update_is_disabled,
)
from ansys.solutions.oscillator_calibration.utilities.common_functions import (
    extract_dict_by_key,
    get_actor_data,
    get_actor_information_data,
    get_log_data,
    get_statistics_data,
)


def layout(
    actor_is_system: bool, project: Oscillator_CalibrationSolution, selected_actor: str, actor_hid: str = None
) -> html.Div:
    """Layout of the summary view."""

    actor_information_data, actor_log_data, actor_statistics_data = _get_data(actor_hid, selected_actor, project)

    information_loading_div = html.Div(id="information_loading_div", style={"display": "none"})
    trigger_initial_data_load_div = html.Div(id="trigger_initial_data_load", style={"display": "none"})

    information_loading_spinner = dcc.Loading(
        type="circle",
        fullscreen=True,
        color=ColorHex.ANSYS_ORANGE.value,
        style={
            "background-color": ColorRgba.TRANSPARENT_GREY.value,
        },
        children=[
            information_loading_div,
            trigger_initial_data_load_div,
        ],
        delay_show=500,
    )
    information_div = html.Div(id="actor_information_table", children=ActorInformationTableAIO(actor_information_data))

    # Assemble UI components
    content = [
        information_loading_spinner,
        html.Br(),
        dbc.Row(
            [
                dbc.Col(information_div, width=12),
            ]
        ),
        dbc.Row(
            [
                dbc.Col(node_control_layout, width=12),
            ]
        ),
    ]

    if actor_is_system:
        system_files_download = dbc.Row(
            [
                dbc.Col(system_files_download_layout, width=12),
            ]
        )
        content.append(system_files_download)

    content.extend(
        [
            dbc.Row(
                [
                    dbc.Col(ActorLogsTableAIO(actor_log_data, aio_id="actor_logs_table"), width=12),
                ]
            ),
            dbc.Row(
                [
                    dbc.Col(
                        html.Div(id="actor_statistics_table", children=ActorStatisticsTableAIO(actor_statistics_data)),
                        width=12,
                    ),
                ]
            ),
            html.Div(id="dummy-output", style={"display": "block"}),
        ]
    )

    # Build layout

    return html.Div(content)


@callback(
    Output("actor_information_table", "children"),
    Output(ActorLogsTableAIO.ids.datatable("actor_logs_table"), "data"),
    Output(ActorLogsTableAIO.ids.datatable("actor_logs_table"), "columns"),
    Output("actor_statistics_table", "children"),
    Output("trigger_initial_data_load", "children"),
    Input("selected_state_dropdown", "value"),
    State("url", "pathname"),
    State("navigation_tree", "treeItemClicked"),
    prevent_initial_call=True,
)
def reset_view_upon_state_change(actor_hid, pathname, tree_item):
    """Update design table."""

    selected_actor = tree_item["id"]

    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    actor_information_data, log_data_frame, actor_statistics_data = _get_data(actor_hid, selected_actor, project)
    log_data_columns = _extract_actor_logs_columns(log_data_frame)

    return (
        ActorInformationTableAIO(actor_information_data),
        log_data_frame.to_dict("records"),
        log_data_columns,
        ActorStatisticsTableAIO(actor_statistics_data),
        "update_required",
    )


@callback(
    Output("information_loading_div", "children"),
    Input("trigger_initial_data_load", "children"),
    State("selected_state_dropdown", "value"),
    State("url", "pathname"),
    State("navigation_tree", "treeItemClicked"),
)
def load_information(initial_data_load_trigger, actor_hid, pathname, tree_item):

    selected_actor = tree_item["id"]
    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    monitoring_step = project.steps.monitoring_step

    if monitoring_step.server_healthy:
        actor_data_files = monitoring_step.actor_data.list_files()
        actor_data = get_actor_data(actor_data_files, selected_actor)

        if len(actor_data.get("states_ids", [])) <= 1:
            # This actor has no or just one, skip
            return no_update

        monitoring_step.actor_to_update = selected_actor
        monitoring_step.hid_to_update = actor_hid
        monitoring_step.update_specific_actor_data()

    return no_update


@callback(
    Output("actor_information_table", "children"),
    Output(ActorLogsTableAIO.ids.datatable("actor_logs_table"), "data"),
    Output(ActorLogsTableAIO.ids.datatable("actor_logs_table"), "columns"),
    Output("actor_statistics_table", "children"),
    Input(WebsocketStreamListeners.ACTOR_DATA.value, "message"),
    State("selected_state_dropdown", "value"),
    State("url", "pathname"),
    State("navigation_tree", "treeItemClicked"),
    State("activate_auto_update", "on"),
    prevent_initial_call=True,
)
def update_view(actor_data_event, actor_hid, pathname, tree_item, auto_update):
    """Update design table."""

    prevent_event_triggered_update_if_auto_update_is_disabled(callback_context, auto_update, actor_data_event)

    selected_actor = tree_item["id"]
    prevent_actor_data_event_update_if_actor_not_updated(actor_data_event, callback_context, selected_actor)

    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)
    actor_information_data, log_data_frame, actor_statistics_data = _get_data(actor_hid, selected_actor, project)
    log_data_columns = _extract_actor_logs_columns(log_data_frame)

    return (
        ActorInformationTableAIO(actor_information_data),
        log_data_frame.to_dict("records"),
        log_data_columns,
        ActorStatisticsTableAIO(actor_statistics_data),
    )


def _get_data(
    actor_hid: str, selected_actor: str, project: Oscillator_CalibrationSolution, actor_is_system=None
) -> Tuple[dict, pd.DataFrame, dict]:
    actor_data_files = _get_actor_data_files(project)
    actor_information_data = _get_actor_information_data_from_files(
        actor_data_files, selected_actor, actor_hid, project, actor_is_system
    )
    log_data_frame = _get_actor_logs_table_data_from_files(actor_data_files, selected_actor)
    actor_statistics_data = get_statistics_data(actor_data_files, selected_actor)
    return actor_information_data, log_data_frame, actor_statistics_data


def _get_actor_data_files(project: Oscillator_CalibrationSolution) -> list[FileReference]:
    monitoring_step = project.steps.monitoring_step
    actor_data_files = monitoring_step.actor_data.list_files()

    return actor_data_files


def _get_actor_information_data_from_files(
    actor_data_files: list[FileReference],
    selected_actor: str,
    actor_hid: str,
    project: Oscillator_CalibrationSolution,
    actor_is_system=None,
) -> dict:
    # Get default empty actor info
    if actor_is_system is None:
        problem_setup_step = project.steps.problem_setup_step
        actor_is_system = _is_node_system(problem_setup_step, selected_actor)

    actor_information_data = datamodel.extract_actor_information_data({}, {}, actor_is_system)
    # Collect actor information data
    if actor_hid:
        actor_information_data_hid = get_actor_information_data(actor_data_files, selected_actor, actor_hid)
        if actor_information_data_hid:
            actor_information_data = actor_information_data_hid

    return actor_information_data


def _get_actor_logs_table_data_from_files(actor_data_files, selected_actor) -> pd.DataFrame:
    actor_log_data = get_log_data(actor_data_files, selected_actor)

    log_data_frame = pd.DataFrame(actor_log_data)

    return log_data_frame


def _extract_actor_logs_columns(log_data_frame: pd.DataFrame) -> dict:
    return [{"name": i, "id": i, "type": "text"} for i in log_data_frame.columns]


def _is_node_system(problem_setup_step: ProblemSetupStep, selected_actor: str) -> bool:
    actor_info = extract_dict_by_key(
        problem_setup_step.osl_project_tree, "uid", selected_actor, expect_unique=True, return_index=False
    )
    node_is_system = actor_info["kind"] == "system"
    return node_is_system
