# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the visualization view."""

from ansys.saf.glow.client.dashclient import DashClient, callback
from dash.exceptions import PreventUpdate
from dash_extensions.enrich import Input, Output, State, callback_context, html
import optislang_dash_lib

from ansys.solutions.oscillator_calibration.datamodel.shared.websocket_streams import WebsocketStreamListeners
from ansys.solutions.oscillator_calibration.solution.definition import Oscillator_CalibrationSolution
from ansys.solutions.oscillator_calibration.ui.utils.websockets_helpers import (
    prevent_event_triggered_update_if_auto_update_is_disabled,
)
from ansys.solutions.oscillator_calibration.utilities.common_functions import (
    find_node_by_uid,
    get_root_node_from_full_project_status_info,
    try_deserialize,
)


def layout(selected_actor: str) -> html.Div:
    """Layout of the visualization view."""

    full_project_status_info = {}

    visualization_component = optislang_dash_lib.VisualizationComponent(
        id="visualization-component",
        project_state=full_project_status_info,
    )

    visualization_div = html.Div(
        id="visualization-component-div",
        children=[visualization_component],
        style={"display": "block"},
    )

    return visualization_div


@callback(
    Output("visualization-component", "project_state"),
    Input(WebsocketStreamListeners.FULL_PROJECT_INFORMATION.value, "message"),
    Input("selected_state_dropdown", "value"),
    State("url", "pathname"),
    State("navigation_tree", "treeItemClicked"),
    State("activate_auto_update", "on"),
)
def update_view(full_project_status_information_updated, actor_hid, pathname, tree_item, auto_update):
    """Update visualization component."""
    prevent_event_triggered_update_if_auto_update_is_disabled(callback_context, auto_update)

    # Get project
    project = DashClient[Oscillator_CalibrationSolution].get_project(pathname)

    # Get monitoring step
    monitoring_step = project.steps.monitoring_step

    full_project_status_info = try_deserialize(monitoring_step.full_project_status_information)

    actor_uid = tree_item["id"]
    system_state = get_system_state(actor_uid, actor_hid, full_project_status_info)
    return system_state


def get_system_state(actor_uid: str, actor_hid: str, full_project_status_info: dict) -> dict:

    try:
        root_node = get_root_node_from_full_project_status_info(full_project_status_info)
    except:
        raise PreventUpdate

    current_node_status_info = find_node_by_uid(root_node, actor_uid)
    if not current_node_status_info:
        raise PreventUpdate

    current_node_status_info_for_hid = next(
        (item for item in current_node_status_info.get("states", []) if "hid" in item and item["hid"] == actor_hid),
        None,
    )
    return current_node_status_info_for_hid
