# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Common functions."""

from collections.abc import Iterable
from functools import partial
import json
import logging
import os
from pathlib import Path, PureWindowsPath
import re
from typing import Any, Callable, Iterable, List, Union

from ansys.optislang.core.osl_server import OslServer
from ansys.saf.glow._core.client_exceptions import ConflictException
from ansys.saf.glow.core.method_status import MethodStatus
from ansys.saf.glow.solution import StepModel
from retrying import retry

from ansys.solutions.oscillator_calibration.datamodel.osl.models import OslTreeItem
from ansys.solutions.oscillator_calibration.datamodel.shared.models import AlertLevel
from ansys.solutions.oscillator_calibration.datamodel.ui.models import TreeviewItem
from ansys.solutions.oscillator_calibration.ui.utils.icons import Icon

LOG_MESSAGE_COLORS = {AlertLevel.INFO: "primary", AlertLevel.WARNING: "warning", AlertLevel.ERROR: "danger"}


def find_dicts_by_key_recursively(structure, target_level, current_path=[]):
    results = []

    if isinstance(structure, list):
        for i, item in enumerate(structure):
            path = current_path + [i]
            results.extend(find_dicts_by_key_recursively(item, target_level, path))
    elif isinstance(structure, dict):
        if "level" in structure and structure["level"] == target_level:
            results.append(current_path)
        for key, value in structure.items():
            path = current_path + [key]
            results.extend(find_dicts_by_key_recursively(value, target_level, path))

    return results


def get_dict_from_indexes_sequence(nested_structure, mixed_keys):
    """
    Retrieve a dictionary from a complex nested structure of dictionaries and lists
    using a mixed list of keys and indexes.

    Args:
        nested_structure (dict or list): The complex nested structure.
        mixed_keys (list): List of keys and indexes to follow, in mixed order.

    Returns:
        dict: The retrieved dictionary, or None if not found.
    """
    for key_or_index in mixed_keys:
        if isinstance(nested_structure, dict) and isinstance(key_or_index, str) and key_or_index in nested_structure:
            nested_structure = nested_structure[key_or_index]
        elif (
            isinstance(nested_structure, list)
            and isinstance(key_or_index, int)
            and 0 <= key_or_index < len(nested_structure)
        ):
            nested_structure = nested_structure[key_or_index]
        else:
            return None  # Key or index not found in the current level

    if isinstance(nested_structure, dict):
        return nested_structure
    else:
        return None  # If the final element is not a dictionary


def get_treeview_items_from_project_tree(osl_project_tree: list[dict]) -> list[dict]:

    treeview_items = [TreeviewItem(id="problem_setup_step", text="Problem Setup", icon=Icon.SETTINGS, level=0)]

    if not osl_project_tree:
        return [treeview_items[0].to_treeview_item_dict()]

    osl_project_tree_converted = [OslTreeItem(**item) for item in osl_project_tree]

    root_node_candidates = [item for item in osl_project_tree_converted if item.is_root == True]
    if len(root_node_candidates) != 1:
        raise Exception(f"Could not find unique root node in osl project tree.")
    treeview_items.append(TreeviewItem.from_osl_tree_item(root_node_candidates[0]))

    previous_item = treeview_items[-1]

    for node in osl_project_tree_converted:
        if node.is_root:
            continue

        node_treeview_item = TreeviewItem.from_osl_tree_item(node)
        parent = _find_parent(parent_candidate=previous_item, current_item=node_treeview_item)

        parent.add_child(node_treeview_item)
        previous_item = node_treeview_item

    return [item.to_treeview_item_dict() for item in treeview_items]


def _find_parent(parent_candidate: TreeviewItem, current_item: TreeviewItem):
    parent_level = current_item.level - 1
    while parent_candidate.level != parent_level:
        if parent_candidate.parent is None:
            name = current_item.text
            raise Exception(f"Unable to find parent node for node {name}.")

        parent_candidate = parent_candidate.parent
    return parent_candidate


def check_empty_strings(lst) -> bool:
    for sublist in lst:
        for item in sublist:
            if not item.strip():  # Using strip() to remove leading/trailing whitespaces
                return False
    return True


def sort_dict_by_ordered_keys(dictionary: dict, list_of_keys: list) -> dict:
    return {key: dictionary[key] for key in list_of_keys if key in dictionary}


def sorted_nicely(l) -> Iterable:
    """Sort the given iterable in the way that humans expect.
    Reference: https://stackoverflow.com/questions/2669059/how-to-sort-alpha-numeric-set-in-python
    """

    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda key: [convert(c) for c in re.split("([0-9]+)", key)]
    return sorted(l, key=alphanum_key)


def convert_microseconds(duration: int) -> str:
    """Convert a time duration in microseconds into a string with format: WWh XXmin YYs ZZms."""

    milliseconds = duration // 1000
    microseconds = duration % 1000
    seconds = milliseconds // 1000
    milliseconds = milliseconds % 1000
    minutes = seconds // 60
    seconds = seconds % 60
    hours = minutes // 60
    minutes = minutes % 60

    return f"{int(hours):02}h {int(minutes):02}m {int(seconds):02}s {int(milliseconds):03}ms"


def extract_dict_by_key(
    dictionaries: list, key: str, value: Any, expect_unique: bool = False, return_index: bool = True
) -> Union[None, dict]:
    """Given a list of dictionaries, return the ones matching specific key and value."""

    matches, indexes = [], []

    for index, dictionary in enumerate(dictionaries):
        if dictionary.get(key) == value:
            matches.append(dictionary)
            indexes.append(index)

    if expect_unique:
        if len(matches) == 0:
            raise Exception(f"No matching dictionaries with key {key} and value {value}.")
        elif len(matches) == 1:
            result_m = matches[0]
            result_i = indexes[0]
        else:
            raise Exception(f"Multiple matches found with key {key} and value {value}.")
    else:
        result_m = matches
        result_i = indexes

    return (result_m, result_i) if return_index else result_m


def read_log_file(log_file: Union[Path, str]) -> list:

    logs = []

    log_file = Path(log_file)

    if log_file.exists():
        with open(log_file, "r") as file:
            for line in file:
                if len(line.strip().rstrip("\n").replace(" ", "")) > 0:
                    logs.append(line.strip().rstrip("\n"))
    else:
        raise FileNotFoundError(f"Unable to find log file {log_file}.")

    return logs


def update_placeholders(ui_values: list, placeholders: dict) -> dict:
    """Information needed."""

    updated_dict = {}
    for row in ui_values[1:]:
        parameter_name = row["props"]["children"][0]["props"]["children"]["props"]["children"]
        input_value = row["props"]["children"][1]["props"]["children"]["props"]["value"]

        placeholder_values = placeholders.get("placeholder_values")
        if parameter_name in placeholder_values:
            updated_dict[parameter_name] = input_value
    return updated_dict


def check_optislang_server(osl_server: OslServer) -> None:
    """optiSLang server health check."""

    try:
        server_is_alive = osl_server.get_server_is_alive()
    except Exception as e:
        return False

    return server_is_alive


def milliseconds(seconds: float) -> float:
    return 1000 * seconds


def seconds(milliseconds: float) -> float:
    return milliseconds / 1000


def print_wait(attempts: int, delay: float):
    # This function is only here to get a log for every retry
    logger = logging.getLogger(__name__)
    logger.debug(f"Attempt #{attempts}, retrying in {seconds(milliseconds=delay)} seconds")
    return delay


retry_strategy = partial(
    retry,
    stop_max_attempt_number=8,
    wait_exponential_multiplier=milliseconds(seconds=0.1),
    wait_exponential_max=milliseconds(seconds=2),
    wait_func=print_wait,
)


@retry_strategy()
def try_deserialize(path: Path) -> dict:
    """This function will try to deserialize the text at the given path.
    It will retry with exponential backoff on any exception, up to 8 times.
    """
    serialized = path.read_text()
    return json.loads(serialized)


@retry_strategy()
def try_read_text(path: Path) -> str:
    """This function will try to read the text at the given path.
    It will retry with exponential backoff on any exception, up to 8 times.
    """
    text = path.read_text()
    return text


def _get_actor_file_data(file_references: List[Path], selected_actor: str, filename: str) -> dict:
    data_file = next(iter(f for f in file_references if f"{selected_actor}/{filename}" in f.url), None)

    if data_file:
        actor_data = try_deserialize(data_file)
    else:
        actor_data = {}

    return actor_data


def get_actor_data(file_references: List[Path], selected_actor: str) -> dict:
    return _get_actor_file_data(file_references, selected_actor, "actor_data.json")


def get_log_data(file_references: List[Path], selected_actor: str) -> dict:
    return _get_actor_file_data(file_references, selected_actor, "log.json")


def get_statistics_data(file_references: List[Path], selected_actor: str) -> dict:
    return _get_actor_file_data(file_references, selected_actor, "statistics.json")


def get_actor_information_data(file_references: List[Path], selected_actor: str, selected_hid: str) -> dict:
    actor_information_file = next(
        iter(f for f in file_references if f"{selected_actor}/{selected_hid}/information.json" in f.url), None
    )

    if actor_information_file:
        actor_information = try_deserialize(actor_information_file)
    else:
        actor_information = {}

    return actor_information


def get_design_table_data(file_references: List[Path], selected_actor: str, selected_hid: str) -> dict:
    design_table_file = next(
        iter(f for f in file_references if f"{selected_actor}/{selected_hid}/design_table.json" in f.url), None
    )

    if design_table_file:
        actor_information = try_deserialize(design_table_file)
    else:
        actor_information = {"Design": []}

    return actor_information


def get_omdb_files_data(file_references: List[Path], selected_actor: str, selected_hid: str) -> dict:
    omdb_files_file = next(
        iter(f for f in file_references if f"{selected_actor}/{selected_hid}/omdb_files.json" in f.url), None
    )

    if omdb_files_file:
        actor_information = try_deserialize(omdb_files_file)
    else:
        actor_information = {}

    return actor_information


def start_transaction_method_if_not_running(step: StepModel, method_name: str, method: Callable[[], None]) -> None:
    logger = logging.getLogger(__name__)

    state = step.get_method_state(method_name)
    status = state.status
    if status == MethodStatus.Failed:
        logger.warning(state.exception_message)
        logger.warning(state.exception_stack)

    if status != MethodStatus.Running:
        try:
            method()
            logger.debug(f"Started {method_name} since state was {status}")
        except ConflictException:
            logger.debug(f"Got conflict exception starting {method_name}. Reported status was {status}")


def get_root_node_from_full_project_status_info(full_project_status_info: dict) -> dict:
    return full_project_status_info["projects"][0]["system"]


def find_node_by_uid(node: dict, uid_to_find: str) -> Union[None, dict]:
    if node["uid"] == uid_to_find:
        return node

    for child in node.get("nodes", []):
        result = find_node_by_uid(child, uid_to_find)
        if result:
            return result

    return None


def convert_path(path: Path) -> Path:
    if path.exists():
        return path
    try:
        mnt_volume = os.getenv("CONTAINER_MNT_VOLUME")

        if not mnt_volume:
            return path

        win_path = PureWindowsPath(str(path))
        drive = win_path.drive  # For example C:

        lin_path = Path(mnt_volume) / drive[0].lower()  # For example /mnt/c
        lin_path = lin_path.joinpath(*win_path.parts[1:])  # Join /mnt/c with rest of parts

        if lin_path.exists():
            return lin_path
        return path
    except:
        return path
